import io
import lzma
import os

import numpy as np
import pytest

import asdf
from asdf import _compression, config_context, generic_io
from asdf._tests import _helpers as helpers
from asdf.extension import Compressor, Extension

RNG = np.random.default_rng(0)


def _get_large_tree():
    x = RNG.normal(size=(128, 128))
    return {"science_data": x}


def _get_sparse_tree():
    arr = np.zeros((128, 128))
    for x, y, z in RNG.normal(size=(64, 3)):
        arr[int(x * 127), int(y * 127)] = z
    arr[0, 0] = 5.0
    return {"science_data": arr}


def _roundtrip(tmp_path, tree, compression=None, write_options=None, read_options=None):
    read_options = {} if read_options is None else read_options
    write_options = {} if write_options is None else write_options.copy()
    write_options.update(all_array_compression=compression)

    tmpfile = os.path.join(str(tmp_path), "test.asdf")

    ff = asdf.AsdfFile(tree)
    ff.write_to(tmpfile, **write_options)

    with asdf.open(tmpfile, mode="rw") as ff:
        ff.update(**write_options)

    with asdf.open(tmpfile, **read_options) as ff:
        helpers.assert_tree_match(tree, ff.tree)

    # Also test saving to a buffer
    buff = io.BytesIO()

    ff = asdf.AsdfFile(tree)
    ff.write_to(buff, **write_options)

    buff.seek(0)
    with asdf.open(buff, **read_options) as ff:
        helpers.assert_tree_match(tree, ff.tree)

    # Test saving to a non-seekable buffer
    buff = io.BytesIO()

    ff = asdf.AsdfFile(tree)
    ff.write_to(generic_io.OutputStream(buff), **write_options)

    buff.seek(0)
    with asdf.open(generic_io.InputStream(buff), **read_options) as ff:
        helpers.assert_tree_match(tree, ff.tree)

    return tmpfile


def test_invalid_compression():
    tree = _get_large_tree()
    ff = asdf.AsdfFile(tree)
    with pytest.raises(ValueError, match=r"Invalid compression type: foo"):
        ff.set_array_compression(tree["science_data"], "foo")
    with pytest.raises(ValueError, match=r"Unknown compression type: .*"):
        _compression._get_compressor("foo")


def test_get_compressed_size():
    assert _compression.get_compressed_size(b"0" * 1024, "zlib") < 1024


def test_decompress_too_long_short():
    fio = io.BytesIO()
    _compression.compress(fio, b"0" * 1024, "zlib")
    size = fio.tell()
    fio.seek(0)
    blocks = lambda us: [fio.read(us)]  # noqa: E731
    fio.read_blocks = blocks
    _compression.decompress(fio, size, 1024, "zlib")
    fio.seek(0)
    with pytest.raises(ValueError, match=r"Decompressed data wrong size"):
        _compression.decompress(fio, size, 1025, "zlib")
    fio.seek(0)
    with pytest.raises(ValueError, match=r"memoryview assignment: lvalue and rvalue have different structures"):
        _compression.decompress(fio, size, 1023, "zlib")


def test_zlib(tmp_path):
    tree = _get_large_tree()

    _roundtrip(tmp_path, tree, "zlib")


def test_bzp2(tmp_path):
    tree = _get_large_tree()

    _roundtrip(tmp_path, tree, "bzp2")


def test_lz4(tmp_path):
    pytest.importorskip("lz4")
    tree = _get_large_tree()

    _roundtrip(tmp_path, tree, "lz4")


def test_recompression(tmp_path):
    tree = _get_large_tree()
    tmpfile = os.path.join(str(tmp_path), "test1.asdf")
    afile = asdf.AsdfFile(tree)
    afile.write_to(tmpfile, all_array_compression="zlib")
    afile.close()
    afile = asdf.open(tmpfile)
    tmpfile = os.path.join(str(tmp_path), "test2.asdf")
    afile.write_to(tmpfile, all_array_compression="bzp2")
    afile.close()
    afile = asdf.open(tmpfile)
    helpers.assert_tree_match(tree, afile.tree)
    afile.close()


def test_input(tmp_path):
    tree = _get_large_tree()
    tmpfile = os.path.join(str(tmp_path), "test1.asdf")
    afile = asdf.AsdfFile(tree)
    afile.write_to(tmpfile, all_array_compression="zlib")
    afile.close()
    afile = asdf.open(tmpfile)
    tmpfile = os.path.join(str(tmp_path), "test2.asdf")
    afile.write_to(tmpfile)
    afile.close()
    afile = asdf.open(tmpfile)
    helpers.assert_tree_match(tree, afile.tree)
    assert afile.get_array_compression(afile.tree["science_data"]) == "zlib"
    afile.close()


def test_none(tmp_path):
    tree = _get_large_tree()

    tmpfile1 = os.path.join(str(tmp_path), "test1.asdf")
    with asdf.AsdfFile(tree) as afile:
        afile.write_to(tmpfile1)

    tmpfile2 = os.path.join(str(tmp_path), "test2.asdf")
    with asdf.open(tmpfile1) as afile:
        assert afile.get_array_compression(afile.tree["science_data"]) is None
        afile.write_to(tmpfile2, all_array_compression="zlib")

    with asdf.open(tmpfile2) as afile:
        assert afile.get_array_compression(afile.tree["science_data"]) == "zlib"
        afile.write_to(tmpfile1, all_array_compression=None)

    with asdf.open(tmpfile1) as afile:
        helpers.assert_tree_match(tree, afile.tree)
        assert afile.get_array_compression(afile.tree["science_data"]) is None


def test_set_array_compression(tmp_path):
    tmpfile = os.path.join(str(tmp_path), "compressed.asdf")

    zlib_data = np.array(list(range(1000)))
    bzp2_data = np.array(list(range(1000)))

    tree = {"zlib_data": zlib_data, "bzp2_data": bzp2_data}
    with asdf.AsdfFile(tree) as af_out:
        af_out.set_array_compression(zlib_data, "zlib", level=1)
        af_out.set_array_compression(bzp2_data, "bzp2", compresslevel=9000)
        with pytest.raises(ValueError, match=r"compresslevel must be between 1 and 9"):
            af_out.write_to(tmpfile)
        af_out.set_array_compression(bzp2_data, "bzp2", compresslevel=9)
        af_out.write_to(tmpfile)
        assert af_out.get_array_compression_kwargs(bzp2_data)["compresslevel"] == 9

    with asdf.open(tmpfile) as af_in:
        assert af_in.get_array_compression(af_in.tree["zlib_data"]) == "zlib"
        assert af_in.get_array_compression(af_in.tree["bzp2_data"]) == "bzp2"


def test_nonnative_endian_compression(tmp_path):
    ledata = np.arange(1000, dtype="<i8")
    bedata = np.arange(1000, dtype=">i8")

    _roundtrip(tmp_path, {"ledata": ledata, "bedata": bedata}, "lz4")


class LzmaCompressor(Compressor):
    def compress(self, data, **kwargs):
        comp = lzma.compress(data, **kwargs)
        yield comp

    def decompress(self, blocks, out, **kwargs):
        decompressor = lzma.LZMADecompressor(**kwargs)
        i = 0
        for block in blocks:
            decomp = decompressor.decompress(block)
            out[i : i + len(decomp)] = decomp
            i += len(decomp)
        return i

    @property
    def label(self):
        return b"lzma"


class LzmaExtension(Extension):
    @property
    def extension_uri(self):
        return "asdf://somewhere.org/extensions/lzma-1.0"

    @property
    def compressors(self):
        return [LzmaCompressor()]


def test_compression_with_extension(tmp_path):
    tree = _get_large_tree()

    with pytest.raises(ValueError, match="Supported compression types are"), config_context() as cfg:
        cfg.all_array_compression = "lzma"

    with config_context() as config:
        config.add_extension(LzmaExtension())

        with config_context() as cfg:
            cfg.all_array_compression = "lzma"

        with pytest.raises(lzma.LZMAError, match=r"Invalid or unsupported options"):
            _roundtrip(tmp_path, tree, "lzma", write_options={"compression_kwargs": {"preset": 9000}})
        fn = _roundtrip(tmp_path, tree, "lzma", write_options={"compression_kwargs": {"preset": 6}})

        hist = {
            "extension_class": "asdf._tests.test_compression.LzmaExtension",
            "extension_uri": "asdf://somewhere.org/extensions/lzma-1.0",
            "supported_compression": ["lzma"],
        }

        with asdf.open(fn) as af:
            assert hist in af["history"]["extensions"]



from __future__ import annotations

import numbers
import typing

import attr

from asdf._jsonschema.exceptions import UndefinedTypeCheck


def _as_dict(init):
    init = init or {}
    return dict(init)


def is_array(checker, instance):
    return isinstance(instance, list)


def is_bool(checker, instance):
    return isinstance(instance, bool)


def is_integer(checker, instance):
    # bool inherits from int, so ensure bools aren't reported as ints
    if isinstance(instance, bool):
        return False
    return isinstance(instance, int)


def is_null(checker, instance):
    return instance is None


def is_number(checker, instance):
    # bool inherits from int, so ensure bools aren't reported as ints
    if isinstance(instance, bool):
        return False
    return isinstance(instance, numbers.Number)


def is_object(checker, instance):
    return isinstance(instance, dict)


def is_string(checker, instance):
    return isinstance(instance, str)


def is_any(checker, instance):
    return True


@attr.s(frozen=True, repr=False)
class TypeChecker:
    """
    A :kw:`type` property checker.

    A `TypeChecker` performs type checking for a `Validator`, converting
    between the defined JSON Schema types and some associated Python types or
    objects.

    Modifying the behavior just mentioned by redefining which Python objects
    are considered to be of which JSON Schema types can be done using
    `TypeChecker.redefine` or `TypeChecker.redefine_many`, and types can be
    removed via `TypeChecker.remove`. Each of these return a new `TypeChecker`.

    Arguments:

        type_checkers:

            The initial mapping of types to their checking functions.
    """

    _type_checkers = attr.ib(default={}, converter=_as_dict)

    def __repr__(self):
        types = ", ".join(repr(k) for k in sorted(self._type_checkers))
        return f"<{self.__class__.__name__} types={{{types}}}>"

    def is_type(self, instance, type: str) -> bool:
        """
        Check if the instance is of the appropriate type.

        Arguments:

            instance:

                The instance to check

            type:

                The name of the type that is expected.

        Raises:

            `asdf._jsonschema.exceptions.UndefinedTypeCheck`:

                if ``type`` is unknown to this object.
        """
        try:
            fn = self._type_checkers[type]
        except KeyError:
            raise UndefinedTypeCheck(type) from None

        return fn(self, instance)

    def redefine(self, type: str, fn) -> "TypeChecker":
        """
        Produce a new checker with the given type redefined.

        Arguments:

            type:

                The name of the type to check.

            fn (collections.abc.Callable):

                A callable taking exactly two parameters - the type
                checker calling the function and the instance to check.
                The function should return true if instance is of this
                type and false otherwise.
        """
        return self.redefine_many({type: fn})

    def redefine_many(self, definitions=()) -> "TypeChecker":
        """
        Produce a new checker with the given types redefined.

        Arguments:

            definitions (dict):

                A dictionary mapping types to their checking functions.
        """
        type_checkers = self._type_checkers.copy()
        type_checkers.update(definitions)
        return attr.evolve(self, type_checkers=type_checkers)

    def remove(self, *types) -> "TypeChecker":
        """
        Produce a new checker with the given types forgotten.

        Arguments:

            types:

                the names of the types to remove.

        Raises:

            `asdf._jsonschema.exceptions.UndefinedTypeCheck`:

                if any given type is unknown to this object
        """

        type_checkers = self._type_checkers.copy()
        for each in types:
            try:
                del type_checkers[each]
            except KeyError:
                raise UndefinedTypeCheck(each)
        return attr.evolve(self, type_checkers=type_checkers)


draft3_type_checker = TypeChecker(
    {
        "any": is_any,
        "array": is_array,
        "boolean": is_bool,
        "integer": is_integer,
        "object": is_object,
        "null": is_null,
        "number": is_number,
        "string": is_string,
    },
)
draft4_type_checker = draft3_type_checker.remove("any")
draft6_type_checker = draft4_type_checker.redefine(
    "integer",
    lambda checker, instance: (
        is_integer(checker, instance)
        or isinstance(instance, float) and instance.is_integer()
    ),
)
draft7_type_checker = draft6_type_checker
draft201909_type_checker = draft7_type_checker
draft202012_type_checker = draft201909_type_checker



import contextlib
import io
import re
from collections import OrderedDict

import numpy as np
import pytest
import yaml

import asdf
from asdf import tagged, treeutil, yamlutil
from asdf.exceptions import AsdfSerializationError
from asdf.testing.helpers import yaml_to_asdf


def _roundtrip(obj, init_kwargs=None):
    """
    Parameters
    ----------

    obj : object
        object to write to ASDF file (under key 'obj')

    Returns
    -------

    file_contents: bytes
        contents of written file

    read_tree : object
        object read back from ASDF file
    """

    init_kwargs = init_kwargs or {}
    buff = io.BytesIO()
    af = asdf.AsdfFile(**init_kwargs)
    af["obj"] = obj
    af.write_to(buff)
    buff.seek(0)

    open_kwargs = {
        "lazy_load": False,
        "memmap": False,
    }

    with asdf.open(buff, **open_kwargs) as af:
        return buff.getvalue(), af["obj"]


def test_ordered_dict():
    """
    Test that we can write out and read in ordered dicts.
    """

    input_tree = {
        "ordered_dict": OrderedDict([("first", "foo"), ("second", "bar"), ("third", "baz")]),
        "unordered_dict": {"first": "foo", "second": "bar", "third": "baz"},
    }

    content, tree = _roundtrip(input_tree)

    assert b"OrderedDict" not in content

    assert isinstance(tree["ordered_dict"], OrderedDict)
    assert tree["ordered_dict"] == input_tree["ordered_dict"]

    assert not isinstance(tree["unordered_dict"], OrderedDict)
    assert isinstance(tree["unordered_dict"], dict)
    assert tree["unordered_dict"] == input_tree["unordered_dict"]


def test_unicode_write():
    """
    We want to write unicode out as regular utf-8-encoded
    characters, not as escape sequences
    """

    input_tree = {"ɐʇɐp‾ǝpoɔıun": 42, "ascii_only": "this is ascii"}  # noqa: RUF001

    content, tree = _roundtrip(input_tree)

    # Ensure that unicode is written out as UTF-8 without escape
    # sequences
    assert "ɐʇɐp‾ǝpoɔıun".encode() in content  # noqa: RUF001
    # Ensure that the unicode "tag" is not used
    assert b"unicode" not in content

    assert tree["ɐʇɐp‾ǝpoɔıun"] == input_tree["ɐʇɐp‾ǝpoɔıun"]  # noqa: RUF001
    assert isinstance(tree["ascii_only"], str)
    assert tree["ascii_only"] == input_tree["ascii_only"]


def test_arbitrary_python_object():
    """
    Putting "just any old" Python object in the tree should raise an
    exception.
    """

    class Foo:
        pass

    tree = {"object": Foo()}

    buff = io.BytesIO()
    ff = asdf.AsdfFile(tree)
    with pytest.raises(AsdfSerializationError, match=r".*is not serializable by asdf.*"):
        ff.write_to(buff)


def test_python_tuple():
    """
    We don't want to store tuples as tuples, because that's not a
    built-in YAML data type.  This test ensures that they are
    converted to lists.
    """

    input_tree = {"val": (1, 2, 3)}

    content, tree = _roundtrip(input_tree)

    assert b"tuple" not in content
    assert isinstance(tree["val"], list)


@contextlib.contextmanager
def multi_warn(category, matches):
    with pytest.warns(category) as record:
        yield

    for match in matches:
        found = False
        for r in record:
            msg = str(r.message)
            if re.match(match, msg):
                found = True
                break
        assert found, f"Did not raise {category} with message matching {match}"


@pytest.mark.xfail(reason="pyyaml has a bug and does not support tuple keys")
def test_python_tuple_key():
    """
    This tests whether tuple keys are round-tripped properly.

    As of this writing, this does not work in pyyaml but does work in
    ruamel.yaml. If/when we decide to switch to ruamel.yaml, this test should
    pass.
    """
    input_tree = {(42, 1): "foo"}

    _, tree = _roundtrip(input_tree)
    assert tree[(42, 1)] == "foo"


def test_tags_removed_after_load(tmp_path):
    input_tree = {"foo": ["bar", (1, 2, None)]}

    _, tree = _roundtrip(input_tree)

    for node in treeutil.iter_tree(tree):
        if node != tree:
            assert not isinstance(node, tagged.Tagged)


def test_explicit_tags():
    yaml = b"""#ASDF 1.0.0
#ASDF_STANDARD 1.5.0
%YAML 1.1
--- !<tag:stsci.edu:asdf/core/asdf-1.1.0>
foo: !<tag:stsci.edu:asdf/core/ndarray-1.0.0> [1, 2, 3]
..."""

    # Check that fully qualified explicit tags work
    buff = io.BytesIO(yaml)

    with asdf.open(buff) as ff:
        assert all(ff.tree["foo"] == [1, 2, 3])


def test_yaml_internal_reference(tmp_path):
    """
    Test that YAML internal references (anchors and aliases) work,
    as well as recursive data structures.
    """

    d = {
        "foo": "2",
    }
    d["bar"] = d

    _list = []
    _list.append(_list)

    input_tree = {"first": d, "second": d, "list": _list}

    content, tree = _roundtrip(input_tree)

    assert b"list:&id002-*id002" in b"".join(content.split())
    assert tree["list"][0] == tree["list"]
    for k in ("first", "second"):
        assert tree[k]["foo"] == input_tree[k]["foo"]
        assert tree[k]["bar"] is tree[k]


def test_yaml_nan_inf():
    tree = {"a": np.nan, "b": np.inf, "c": -np.inf}

    buff = io.BytesIO()
    ff = asdf.AsdfFile(tree)
    ff.write_to(buff)
    buff.seek(0)
    with asdf.open(buff) as ff:
        assert np.isnan(ff.tree["a"])
        assert np.isinf(ff.tree["b"])
        assert np.isinf(ff.tree["c"])


def test_tag_object():
    class SomeObject:
        pass

    tag = "tag:nowhere.org:none/some/thing"
    instance = tagged.tag_object(tag, SomeObject())
    assert instance._tag == tag


@pytest.mark.parametrize(
    ("numpy_value", "expected_value"),
    [
        (np.str_("foo"), "foo"),
        (np.bytes_("foo"), b"foo"),
        (np.float16(3.14), 3.14),
        (np.float32(3.14159), 3.14159),
        (np.float64(3.14159), 3.14159),
        # Evidently float128 is not available on Windows:
        (getattr(np, "float128", np.float64)(3.14159), 3.14159),
        (np.int8(42), 42),
        (np.int16(42), 42),
        (np.int32(42), 42),
        (np.int64(42), 42),
        (np.longlong(42), 42),
        (np.uint8(42), 42),
        (np.uint16(42), 42),
        (np.uint32(42), 42),
        (np.uint64(42), 42),
        (np.ulonglong(42), 42),
    ],
)
def test_numpy_scalar(numpy_value, expected_value):
    ctx = asdf.AsdfFile({"value": numpy_value})
    tree = ctx.tree
    buffer = io.BytesIO()

    yamlutil.dump_tree(tree, buffer, ctx)
    buffer.seek(0)

    loaded_value = yamlutil.load_tree(buffer)["value"]
    if isinstance(numpy_value, np.floating):
        abs_diff = abs(expected_value - loaded_value)
        eps = np.finfo(numpy_value.dtype).eps
        assert abs_diff < eps, abs_diff
    else:
        assert loaded_value == expected_value


@pytest.mark.parametrize(
    "payload",
    [
        "  1: a",  # not a sequence
        "- !!omap\n  - 1",  # sequence item, not a mapping
        "- !!omap\n  1: a\n  2: a",  # sequence item, not a one element mapping
    ],
)
def test_invalid_omap(payload):
    test_yaml = f"""od: !!omap
{payload}
"""

    # Check that fully qualified explicit tags work
    buff = yaml_to_asdf(test_yaml)

    with pytest.raises(yaml.constructor.ConstructorError):
        with asdf.open(buff) as ff:
            ff["od"]



import re

import numpy as np

from asdf import util
from asdf.extension import Converter

_REPLACEMENTS = {
    re.compile("i(?!nf)"): "j",
    re.compile("I(?!NF)"): "J",
}


class ComplexConverter(Converter):
    tags = ["tag:stsci.edu:asdf/core/complex-1.0.0"]
    types = [*list(util._iter_subclasses(np.complexfloating)), complex]

    def to_yaml_tree(self, obj, tag, ctx):
        return str(obj)

    def from_yaml_tree(self, node, tag, ctx):
        for pattern, replacement in _REPLACEMENTS.items():
            node = pattern.sub(replacement, node)

        return complex(node)



import numpy as np
import pytest

import asdf
from asdf._block import external


def test_cache(tmp_path):
    efn = tmp_path / "test.asdf"
    arr = np.arange(3, dtype="uint8")
    asdf.AsdfFile({"data": arr}).write_to(efn)

    cache = external.ExternalBlockCache()
    base_uri = f"{tmp_path.as_uri()}/"
    data = cache.load(base_uri, "test.asdf")
    np.testing.assert_array_equal(data, arr)
    assert cache.load(base_uri, "test.asdf") is data
    assert cache.load(base_uri, "#") is external.UseInternal
    assert cache.load(base_uri, "") is external.UseInternal


@pytest.mark.parametrize("uri", ["test.asdf", "foo/test.asdf"])
@pytest.mark.parametrize("index", [0, 1, 100])
def test_relative_uri_for_index(uri, index):
    match = f"test{index:04d}.asdf"
    assert external.relative_uri_for_index(uri, index) == match



import collections
import contextlib
import copy

from asdf import config, constants, generic_io, util

from . import external, reader, store, writer
from . import io as bio
from .callback import DataCallback
from .key import Key as BlockKey
from .options import Options


class ReadBlocks(collections.UserList):
    """
    A list of ReadBlock instances.

    A simple list can't be used as other code will need
    to generate a weakref to instances of this class
    (and it is not possible to generate a weakref to a list).
    """

    pass


class WriteBlocks(collections.abc.Sequence):
    """
    A collection of ``WriteBlock`` instances that can be accessed by:
        - numerical index (see ``collections.abc.Sequence``)
        - the object or objects in the tree that created or
          are associated with this block
        - the block data
    Access by object and data is via a Store which generates
    Keys to allow use of non-hashable objects (and to not hold
    a reference to the block data).
    """

    def __init__(self, blocks=None):
        if blocks is None:
            blocks = []
        self._blocks = blocks

        # both stores contain values that are indices of
        # WriteBlock instances in _blocks
        self._data_store = store.Store()
        self._object_store = store.Store()

    def __getitem__(self, index):
        return self._blocks.__getitem__(index)

    def __len__(self):
        return self._blocks.__len__()

    def index_for_data(self, data):
        return self._data_store.lookup_by_object(data)

    def assign_object_to_index(self, obj, index):
        self._object_store.assign_object(obj, index)

    def object_keys_for_index(self, index):
        yield from self._object_store.keys_for_value(index)

    def append_block(self, blk, obj):
        """
        Append a ``WriteBlock`` instance to this collection
        assign an object, obj, to the block and return
        the index of the block within the collection.
        """
        index = len(self._blocks)
        self._blocks.append(blk)

        # assign the block data to this block to allow
        # fast lookup of blocks based on data
        self._data_store.assign_object(blk._data, index)

        # assign the object that created/uses this block
        self._object_store.assign_object(obj, index)
        return index


class OptionsStore(store.Store):
    """
    A ``Store`` of ``Options`` that can be accessed by the base
    array that corresponds to a block. A ``Store`` is used
    to avoid holding references to the array data
    (see ``asdf._block.store.Store``).

    When ``Options`` are not found within the ``Store``, the
    ``OptionsStore`` will look for any available matching
    ``ReadBlock`` to determine default Options.
    """

    def __init__(self, read_blocks):
        super().__init__()
        # ReadBlocks are needed to look up default options
        self._read_blocks = read_blocks

    def has_options(self, array):
        """
        Check of Options have been defined for this array
        without falling back to generating Options from
        a ReadBlock.

        Parameters
        ----------
        array : ndarray
            The base of this array (see `asdf.util.get_array_base`) will
            be used to lookup any Options in the Store.

        Returns
        -------
        has_options : bool
            True if Options were previously defined for this array.
        """
        base = util.get_array_base(array)
        return self.lookup_by_object(base) is not None

    def get_options_from_block(self, array):
        """
        Get Options for some array using only settings read from a
        corresponding ReadBlock (one that shares the same base array).
        Any Options defined using previous calls to set_options will
        be ignored (use ``get_options`` if you would like these previously
        set options to be considered).

        Parameters
        ----------
        array : ndarray
            The base of this array (see `asdf.util.get_array_base`) will
            be used to lookup a corresponding ReadBlock.

        Returns
        -------
        options : Options or None
            Options initialized from settings read from a ReadBlock
            or None if no corresponding block was found.
        """
        base = util.get_array_base(array)
        # look up by block with matching _data
        for block in self._read_blocks:
            if block._cached_data is base or block._data is base:
                # init options
                if block.header["flags"] & constants.BLOCK_FLAG_STREAMED:
                    storage_type = "streamed"
                else:
                    storage_type = "internal"
                options = Options(storage_type, block.header["compression"])
                return options
        return None

    def get_options(self, array):
        """
        Get Options for some array using either previously defined
        options (as set by ``set_options``) or settings read from a
        corresponding ReadBlock (one that shares the same base array).

        Note that if no options are found in the Store and options
        are found from a ReadBlock the resulting Options will be added
        to the Store.

        Parameters
        ----------
        array : ndarray
            The base of this array (see `asdf.util.get_array_base`) will
            be used to lookup any Options in the Store.

        Returns
        -------
        options : Options or None
            Options read from the Store or ReadBlocks or None if
            no options were found.
        """
        base = util.get_array_base(array)
        options = self.lookup_by_object(base)
        if options is None:
            options = self.get_options_from_block(base)
            if options is not None:
                self.set_options(base, options)
        if options is None:
            options = Options()
            self.set_options(base, options)
        return options

    def set_options(self, array, options):
        """
        Set Options for an array.

        Parameters
        ----------
        array : ndarray
            The base of this array (see `asdf.util.get_array_base`) will
            be used to add options to the Store.

        options : Options
            The Options to add to the Store for this array.

        Raises
        ------
        ValueError
            If more than one block is set as a streamed block.
        """
        if options.storage_type == "streamed":
            for oid, by_key in self._by_id.items():
                for key, opt in by_key.items():
                    if not key._is_valid():
                        continue
                    if opt.storage_type == "streamed":
                        if opt is options:
                            continue
                        msg = "Can not add second streaming block"
                        raise ValueError(msg)
        base = util.get_array_base(array)
        self.assign_object(base, options)

    def get_output_compressions(self):
        """
        Get all output compression types used for this Store of
        Options.

        Returns
        -------
        compressions : list of bytes
            List of 4 byte compression labels used for this OptionsStore.
        """
        compressions = set()
        cfg = config.get_config()
        if cfg.all_array_compression == "input":
            for blk in self._read_blocks:
                if blk.header["compression"]:
                    compressions.add(blk.header["compression"])
        else:
            compressions.add(cfg.all_array_compression)
        for _, by_key in self._by_id.items():
            for key, opts in by_key.items():
                if not key._is_valid():
                    continue
                if opts.compression:
                    compressions.add(opts.compression)
        return compressions


class Manager:
    """
    ``Manager`` for reading, writing and storing options for ASDF blocks.

    This class does the heavy lifting of allowing ``asdf.AsdfFile`` instances
    to control ASDF blocks. It is responsible for reading and writing blocks
    primarily to maintain some consistency with the previous BlockManager.

    Block ``Options`` control the compression and type of storage for an
    ASDF block (see `asdf.AsdfFile.set_array_storage`,
    `asdf.AsdfFile.set_array_compression`
    `asdf.AsdfFile.set_array_compression` for relevant usage and information).
    These ``Options`` instances are stored and retrieved using the base
    of the array containing the data for an ASDF block. This allows arrays
    that share the same base array (ie views of the same array) to use
    the same ASDF block.

    Reading blocks occurs through use of ``Manager.read`` which will
    create ``ReadBlock`` instances for each read ASDF block. These ``ReadBlock``
    will be used as the source for default ``Options`` for each block
    and ASDF block data can be read using ``DataCallback`` instances.
    These callbacks are used (instead of just accessing blocks by index)
    to allow block reorganization during ``update``.(Note that reading
    of external blocks is special as these are not stored within the
    block section of the ASDF file. These must be explicitly loaded
    using ``Manager._load_external``).

    Writing ASDF blocks occurs through use of ``Manager.write`` which will
    take any queued ``WriteBlocks`` (created via ``Manager.make_write_block``
    and ``Manager.set_streamed_write_block``) and write them out to a file.
    This writing must occur within a ``Manager.write_context`` to allow the
    ``Manager`` to reset any ``Options`` changes that occur during write
    and to clean up the write queue.

    Update-in-place occurs through use of ``Manager.update`` which, like
    ``Manager.write`` must occur within a ``Manager.write_context``. Following
    a ``Manager.update`` the ``ReadBlock`` instances will be replaced with
    the newly written ASDF blocks and any ``DataCallbacks`` will be updated
    to reference the appropriate new ``ReadBlock``.
    """

    def __init__(self, read_blocks=None, uri=None, lazy_load=False, memmap=False, validate_checksums=False):
        if read_blocks is None:
            read_blocks = ReadBlocks([])
        self.options = OptionsStore(read_blocks)

        self._blocks = read_blocks
        self._external_block_cache = external.ExternalBlockCache()
        self._data_callbacks = store.Store()

        self._write_blocks = WriteBlocks()
        self._external_write_blocks = []
        self._streamed_write_block = None
        self._streamed_obj_keys = set()
        self._write_fd = None

        # store the uri of the ASDF file here so that the Manager can
        # resolve and load external blocks without requiring a reference
        # to the AsdfFile instance
        self._uri = uri

        # general block settings
        self._lazy_load = lazy_load
        self._memmap = memmap
        self._validate_checksums = validate_checksums

    def close(self):
        self._external_block_cache.clear()
        self._clear_write()
        for blk in self.blocks:
            blk.close()
        self.options = OptionsStore(self.blocks)

    @property
    def blocks(self):
        """
        Get any ReadBlocks that were read from an ASDF file

        Returns
        -------
        read_blocks : list of ReadBlock
            List of ReadBlock instances created during a call to read
            or update.
        """
        return self._blocks

    @blocks.setter
    def blocks(self, new_blocks):
        if not isinstance(new_blocks, ReadBlocks):
            new_blocks = ReadBlocks(new_blocks)
        self._blocks = new_blocks
        # we propagate these blocks to options so that
        # options lookups can fallback to the new read blocks
        self.options._read_blocks = new_blocks

    def read(self, fd, after_magic=False):
        """
        Read blocks from an ASDF file and update the manager read_blocks.

        Parameters
        ----------
        fd : file or generic_io.GenericIO
            File to read from. Reading starts at the current file position.

        after_magic : bool, optional, default False
            If True, the file pointer is past the block magic bytes of the
            first block.
        """
        self.blocks = reader.read_blocks(
            fd, self._memmap, self._lazy_load, self._validate_checksums, after_magic=after_magic
        )

    def _load_external(self, uri):
        value = self._external_block_cache.load(self._uri, uri, self._memmap, self._validate_checksums)
        if value is external.UseInternal:
            return self.blocks[0].data
        return value

    def _clear_write(self):
        self._write_blocks = WriteBlocks()
        self._external_write_blocks = []
        self._streamed_write_block = None
        self._streamed_obj_keys = set()
        self._write_fd = None

    def _write_external_blocks(self):
        from asdf import AsdfFile

        if self._write_fd is None or self._write_fd.uri is None:
            msg = "Can't write external blocks, since URI of main file is unknown."
            raise ValueError(msg)

        for blk in self._external_write_blocks:
            uri = generic_io.resolve_uri(self._write_fd.uri, blk._uri)
            af = AsdfFile()
            with generic_io.get_file(uri, mode="w") as f:
                af.write_to(f, include_block_index=False)
                writer.write_blocks(f, [blk])

    def make_write_block(self, data, options, obj):
        """
        Make a WriteBlock with data and options and
        associate it with an object (obj).

        Parameters
        ----------
        data : npdarray or callable
            Data to be written to an ASDF block. Can be provided as
            a callable function that when evaluated will return the
            data.
        options : Options or None
            Options instance used to define the ASDF block compression
            and storage type. If None, a new Options instance will
            be created.
        obj : object
            An object in the ASDF tree that will be associated
            with the new WriteBlock so that `AsdfFile.update` can
            map newly created blocks to blocks read from the original
            file.

        Returns
        -------
        block_source : int or str
            The relative uri (str) if an external block was created
            or the index of the block (int) for an internal block.

        Raises
        ------
        ValueError
            If a external block was created without a URI for the main
            file.
        """
        if options is None:
            options = Options()
        if options.storage_type == "external":
            for index, blk in enumerate(self._external_write_blocks):
                if blk._data is data:
                    # this external uri is already ready to go
                    return blk._uri
            # need to set up new external block
            index = len(self._external_write_blocks)
            blk = writer.WriteBlock(data, options.compression, options.compression_kwargs)
            if self._write_fd is not None:
                base_uri = self._write_fd.uri or self._uri
            else:
                base_uri = self._uri
            if base_uri is None:
                msg = "Can't write external blocks, since URI of main file is unknown."
                raise ValueError(msg)
            blk._uri = external.relative_uri_for_index(base_uri, index)
            self._external_write_blocks.append(blk)
            return blk._uri
        # first, look for an existing block
        index = self._write_blocks.index_for_data(data)
        if index is not None:
            self._write_blocks.assign_object_to_index(obj, index)
            return index
        # if no block is found, make a new block
        blk = writer.WriteBlock(data, options.compression, options.compression_kwargs)
        index = self._write_blocks.append_block(blk, obj)
        return index

    def set_streamed_write_block(self, data, obj):
        """
        Create a WriteBlock that will be written as an ASDF
        streamed block.

        Parameters
        ----------
        data : ndarray or callable
            Data to be written to an ASDF block. Can be provided as
            a callable function that when evaluated will return the
            data.
        obj : object
            An object in the ASDF tree that will be associated
            with the new WriteBlock so that `AsdfFile.update` can
            map newly created blocks to blocks read from the original
            file.
        """
        if self._streamed_write_block is not None and data is not self._streamed_write_block.data:
            msg = "Can not add second streaming block"
            raise ValueError(msg)
        if self._streamed_write_block is None:
            self._streamed_write_block = writer.WriteBlock(data)
        self._streamed_obj_keys.add(BlockKey(obj))

    def _get_data_callback(self, index):
        return DataCallback(index, self.blocks)

    def _set_array_storage(self, data, storage):
        options = self.options.get_options(data)
        options.storage_type = storage
        self.options.set_options(data, options)

    def _get_array_storage(self, data):
        return self.options.get_options(data).storage_type

    def _set_array_compression(self, arr, compression, **compression_kwargs):
        # if this is input compression but we already have defined options
        # we need to re-lookup the options based off the block
        if compression == "input" and self.options.has_options(arr):
            from_block_options = self.options.get_options_from_block(arr)
            if from_block_options is not None:
                compression = from_block_options.compression
        options = self.options.get_options(arr)
        options.compression = compression
        options.compression_kwargs = compression_kwargs

    def _get_array_compression(self, arr):
        return self.options.get_options(arr).compression

    def _get_array_compression_kwargs(self, arr):
        return self.options.get_options(arr).compression_kwargs

    def get_output_compressions(self):
        return self.options.get_output_compressions()

    def _set_array_save_base(self, data, save_base):
        options = self.options.get_options(data)
        options.save_base = save_base
        self.options.set_options(data, options)

    def _get_array_save_base(self, data):
        return self.options.get_options(data).save_base

    @contextlib.contextmanager
    def options_context(self):
        """
        Context manager that copies block options on
        entrance and restores the options when exited.
        """
        previous_options = copy.deepcopy(self.options)
        yield
        self.options = previous_options
        self.options._read_blocks = self.blocks

    @contextlib.contextmanager
    def write_context(self, fd, copy_options=True):
        """
        Context manager that copies block options on
        entrance and restores the options when exited.

        Parameters
        ----------
        fd : file or generic_io.GenericIO
            File to be written to. This is required on
            entrance to this context so that any external
            blocks can resolve relative uris.

        copy_options : bool, optional, default True
            Copy options on entrance and restore them on
            exit (See `options_context`).
        """
        self._clear_write()
        self._write_fd = fd
        if copy_options:
            with self.options_context():
                yield
        else:
            yield
        self._clear_write()

    def write(self, pad_blocks, include_block_index):
        """
        Write blocks that were set up during the current
        `write_context`.

        Parameters
        ----------
        pad_blocks : bool, None or float
            If False, add no padding bytes between blocks. If True
            add some default amount of padding. If a float, add
            a number of padding bytes based off a ratio of the data
            size.

        include_block_index : bool
            If True, include a block index at the end of the file.
            If a streamed_block is provided (or the file is not
            seekable) no block index will be written.

        Raises
        ------
        OSError
            If called outside a `write_context`.
        """
        if self._write_fd is None:
            msg = "write called outside of valid write_context"
            raise OSError(msg)
        if len(self._write_blocks) or self._streamed_write_block:
            writer.write_blocks(
                self._write_fd,
                self._write_blocks,
                pad_blocks,
                streamed_block=self._streamed_write_block,
                write_index=include_block_index,
            )
        if len(self._external_write_blocks):
            self._write_external_blocks()

    def update(self, new_tree_size, pad_blocks, include_block_index):
        """
        Perform an update-in-place of ASDF blocks set up during
        a `write_context`.

        Parameters
        ----------
        new_tree_size : int
            Size (in bytes) of the serialized ASDF tree (and any
            header bytes) that will be written at the start of the
            file being updated.

        pad_blocks : bool, None or float
            If False, add no padding bytes between blocks. If True
            add some default amount of padding. If a float, add
            a number of padding bytes based off a ratio of the data
            size.

        include_block_index : bool
            If True, include a block index at the end of the file.
            If a streamed_block is provided (or the file is not
            seekable) no block index will be written.


        Raises
        ------
        OSError
            If called outside a `write_context`.
        """
        if self._write_fd is None:
            msg = "update called outside of valid write_context"
            raise OSError(msg)
        # find where to start writing blocks (either end of new tree or end of last 'free' block)
        last_block = None
        for blk in self.blocks[::-1]:
            if not blk.memmap and (blk._cached_data is not None or not callable(blk._data)):
                continue
            last_block = blk
            break
        if last_block is None:
            new_block_start = new_tree_size
        else:
            new_block_start = max(
                last_block.data_offset + last_block.header["allocated_size"],
                new_tree_size,
            )

        if len(self._external_write_blocks):
            self._write_external_blocks()

        # do we have any blocks to write?
        if len(self._write_blocks) or self._streamed_write_block:
            self._write_fd.seek(new_block_start)
            offsets, headers = writer.write_blocks(
                self._write_fd,
                self._write_blocks,
                pad_blocks,
                streamed_block=self._streamed_write_block,
                write_index=False,  # don't write an index as we will modify the offsets
            )
            new_block_end = self._write_fd.tell()

            # move blocks to start in increments of block_size
            n_bytes = new_block_end - new_block_start
            src, dst = new_block_start, new_tree_size
            block_size = self._write_fd.block_size
            while n_bytes > 0:
                self._write_fd.seek(src)
                bs = self._write_fd.read(min(n_bytes, block_size))
                self._write_fd.seek(dst)
                self._write_fd.write(bs)
                n = len(bs)
                n_bytes -= n
                src += n
                dst += n

            # update offset to point at correct locations
            offsets = [o - (new_block_start - new_tree_size) for o in offsets]

            # write index if no streamed block
            if include_block_index and self._streamed_write_block is None:
                bio.write_block_index(self._write_fd, offsets)

            # map new blocks to old blocks
            new_read_blocks = ReadBlocks()
            for i, (offset, header) in enumerate(zip(offsets, headers)):
                # find all objects that assigned themselves to the write block at index i
                if i == len(self._write_blocks):  # this is a streamed block
                    obj_keys = self._streamed_obj_keys
                else:
                    # find object associated with this write block
                    obj_keys = set(self._write_blocks.object_keys_for_index(i))

                # we have to be lazy here as any current memmap is invalid
                new_read_block = reader.ReadBlock(offset + 4, self._write_fd, self._memmap, True, False, header=header)
                new_read_blocks.append(new_read_block)
                new_index = len(new_read_blocks) - 1

                # update all callbacks
                for obj_key in obj_keys:
                    obj = obj_key._ref()
                    if obj is None:
                        # this object no longer exists so don't both assigning it
                        continue

                    # update data callbacks to point to new block
                    cb = self._data_callbacks.lookup_by_object(obj)
                    if cb is not None:
                        cb._reassign(new_index, new_read_blocks)

            # update read blocks to reflect new state
            self.blocks = new_read_blocks



import mmap
import sys

import numpy as np
from numpy import ma

from asdf import util
from asdf._jsonschema import ValidationError

_STRUCTURED_DATATYPE_KEYS = {"name", "datatype", "byteorder", "shape"}

_datatype_names = {
    "int8": "i1",
    "int16": "i2",
    "int32": "i4",
    "int64": "i8",
    "uint8": "u1",
    "uint16": "u2",
    "uint32": "u4",
    "uint64": "u8",
    "float16": "f2",
    "float32": "f4",
    "float64": "f8",
    "complex64": "c8",
    "complex128": "c16",
    "bool8": "b1",
}


_string_datatype_names = {"ascii": "S", "ucs4": "U"}


def asdf_byteorder_to_numpy_byteorder(byteorder):
    if byteorder == "big":
        return ">"

    if byteorder == "little":
        return "<"

    msg = f"Invalid ASDF byteorder '{byteorder}'"
    raise ValueError(msg)


def asdf_datatype_to_numpy_dtype(datatype, byteorder=None):
    if byteorder is None:
        byteorder = sys.byteorder

    if isinstance(datatype, str) and datatype in _datatype_names:
        datatype = _datatype_names[datatype]
        byteorder = asdf_byteorder_to_numpy_byteorder(byteorder)
        return np.dtype(str(byteorder + datatype))

    if (
        isinstance(datatype, list)
        and len(datatype) == 2
        and isinstance(datatype[0], str)
        and isinstance(datatype[1], int)
        and datatype[0] in _string_datatype_names
    ):
        length = datatype[1]
        byteorder = asdf_byteorder_to_numpy_byteorder(byteorder)
        datatype = str(byteorder) + str(_string_datatype_names[datatype[0]]) + str(length)

        return np.dtype(datatype)

    if isinstance(datatype, dict):
        if "datatype" not in datatype:
            msg = f"Field entry has no datatype: '{datatype}'"
            raise ValueError(msg)

        name = datatype.get("name", "")
        byteorder = datatype.get("byteorder", byteorder)
        shape = datatype.get("shape")
        datatype = asdf_datatype_to_numpy_dtype(datatype["datatype"], byteorder)

        if shape is None:
            return (str(name), datatype)

        return (str(name), datatype, tuple(shape))

    if isinstance(datatype, list):
        datatype_list = []
        for subdatatype in datatype:
            np_dtype = asdf_datatype_to_numpy_dtype(subdatatype, byteorder)
            if isinstance(np_dtype, tuple):
                datatype_list.append(np_dtype)

            elif isinstance(np_dtype, np.dtype):
                datatype_list.append(("", np_dtype))

            else:
                msg = "Error parsing asdf datatype"
                raise RuntimeError(msg)

        return np.dtype(datatype_list)

    msg = f"Unknown datatype {datatype}"
    raise ValueError(msg)


def numpy_byteorder_to_asdf_byteorder(byteorder, override=None):
    if override is not None:
        return override

    if byteorder == "=":
        return sys.byteorder

    if byteorder == "<":
        return "little"

    return "big"


def numpy_dtype_to_asdf_datatype(dtype, include_byteorder=True, override_byteorder=None):
    dtype = np.dtype(dtype)
    if dtype.names is not None:
        fields = []
        for name in dtype.names:
            field = dtype.fields[name][0]
            d = {}
            d["name"] = name
            field_dtype, byteorder = numpy_dtype_to_asdf_datatype(field, override_byteorder=override_byteorder)
            d["datatype"] = field_dtype
            if include_byteorder:
                d["byteorder"] = byteorder
            if field.shape:
                d["shape"] = list(field.shape)
            fields.append(d)
        return fields, numpy_byteorder_to_asdf_byteorder(dtype.byteorder, override=override_byteorder)

    if dtype.subdtype is not None:
        return numpy_dtype_to_asdf_datatype(dtype.subdtype[0], override_byteorder=override_byteorder)

    if dtype.name in _datatype_names:
        return dtype.name, numpy_byteorder_to_asdf_byteorder(dtype.byteorder, override=override_byteorder)

    if dtype.name == "bool":
        return "bool8", numpy_byteorder_to_asdf_byteorder(dtype.byteorder, override=override_byteorder)

    if dtype.name.startswith("string") or dtype.name.startswith("bytes"):
        return ["ascii", dtype.itemsize], "big"

    if dtype.name.startswith("unicode") or dtype.name.startswith("str"):
        return (
            ["ucs4", int(dtype.itemsize / 4)],
            numpy_byteorder_to_asdf_byteorder(dtype.byteorder, override=override_byteorder),
        )

    msg = f"Unknown dtype {dtype}"
    raise ValueError(msg)


def inline_data_asarray(inline, dtype=None):
    # np.asarray doesn't handle structured arrays unless the innermost
    # elements are tuples.  To do that, we drill down the first
    # element of each level until we find a single item that
    # successfully converts to a scalar of the expected structured
    # dtype.  Then we go through and convert everything at that level
    # to a tuple.  This probably breaks for nested structured dtypes,
    # but it's probably good enough for now.  It also won't work with
    # object dtypes, but ASDF explicitly excludes those, so we're ok
    # there.
    if dtype is not None and dtype.fields is not None:

        def find_innermost_match(line, depth=0):
            if not isinstance(line, list) or not len(line):
                msg = "data can not be converted to structured array"
                raise ValueError(msg)
            try:
                np.asarray(tuple(line), dtype=dtype)
            except ValueError:
                return find_innermost_match(line[0], depth + 1)
            else:
                return depth

        depth = find_innermost_match(inline)

        def convert_to_tuples(line, data_depth, depth=0):
            if data_depth == depth:
                return tuple(line)

            return [convert_to_tuples(x, data_depth, depth + 1) for x in line]

        inline = convert_to_tuples(inline, depth)

        return np.asarray(inline, dtype=dtype)

    def handle_mask(inline):
        if isinstance(inline, list):
            if None in inline:
                inline_array = np.asarray(inline)
                nones = np.equal(inline_array, None)
                return np.ma.array(np.where(nones, 0, inline), mask=nones)

            return [handle_mask(x) for x in inline]

        return inline

    inline = handle_mask(inline)

    inline = np.ma.asarray(inline, dtype=dtype)
    if not ma.is_masked(inline):
        return inline.data

    return inline


def numpy_array_to_list(array):
    def tolist(x):
        if isinstance(x, (np.ndarray, NDArrayType)):
            x = x.astype("U").tolist() if x.dtype.char == "S" else x.tolist()

        if isinstance(x, (list, tuple)):
            return [tolist(y) for y in x]

        return x

    def ascii_to_unicode(x):
        # Convert byte string arrays to unicode string arrays, since YAML
        # doesn't handle the former.
        if isinstance(x, list):
            return [ascii_to_unicode(y) for y in x]

        if isinstance(x, bytes):
            return x.decode("ascii")

        return x

    return ascii_to_unicode(tolist(array))


class NDArrayType:
    def __init__(self, source, shape, dtype, offset, strides, order, mask, data_callback=None):
        self._source = source
        self._data_callback = data_callback
        self._array = None
        self._mask = mask

        if isinstance(source, list):
            self._array = inline_data_asarray(source, dtype)
            self._array = self._apply_mask(self._array, self._mask)
            # single element structured arrays can have shape == ()
            # https://github.com/asdf-format/asdf/issues/1540
            if shape is not None and (
                self._array.shape != tuple(shape)
                or (len(shape) and shape[0] == "*" and self._array.shape[1:] != tuple(shape[1:]))
            ):
                msg = "inline data doesn't match the given shape"
                raise ValueError(msg)

        self._shape = shape
        self._dtype = dtype
        self._offset = offset
        self._strides = strides
        self._order = order

    def _make_array(self):
        # If the ASDF file has been updated in-place, then there's
        # a chance that the block's original data object has been
        # closed and replaced.  We need to check here and re-generate
        # the array if necessary, otherwise we risk segfaults when
        # memory mapping.
        if self._array is not None:
            base = util.get_array_base(self._array)
            if isinstance(base, np.memmap) and isinstance(base.base, mmap.mmap):
                # check if the underlying mmap matches the one generated by generic_io
                try:
                    fd = self._data_callback(_attr="_fd")()
                except AttributeError:
                    # external blocks do not have a '_fd' and don't need to be updated
                    fd = None
                if fd is not None:
                    if getattr(fd, "_mmap", None) is not base.base:
                        self._array = None
                    del fd

        if self._array is None:
            if isinstance(self._source, str):
                # we need to keep _source as a str to allow stdatamodels to
                # support AsdfInFits
                data = self._data_callback()
            else:
                # cached data is used here so that multiple NDArrayTypes will all use
                # the same base array
                data = self._data_callback(_attr="cached_data")

            if hasattr(data, "base") and isinstance(data.base, mmap.mmap) and data.base.closed:
                msg = "ASDF file has already been closed. Can not get the data."
                raise OSError(msg)

            # compute shape (streaming blocks have '0' data size in the block header)
            shape = self.get_actual_shape(
                self._shape,
                self._strides,
                self._dtype,
                data.size,
            )
            self._array = np.ndarray(shape, self._dtype, data, self._offset, self._strides, self._order)
            self._array = self._apply_mask(self._array, self._mask)
        return self._array

    def _apply_mask(self, array, mask):
        if isinstance(mask, (np.ndarray, NDArrayType)):
            # Use "mask.view()" here so the underlying possibly
            # memmapped mask array is freed properly when the masked
            # array goes away.
            array = ma.array(array, mask=mask.view())
            return array

        if np.isscalar(mask):
            if np.isnan(mask):
                return ma.array(array, mask=np.isnan(array))

            return ma.masked_values(array, mask)

        return array

    def __array__(self):
        return self._make_array()

    def __repr__(self):
        # repr alone should not force loading of the data
        if self._array is None:
            return (
                f"<{'array' if self._mask is None else 'masked array'} "
                f"(unloaded) shape: {self._shape} dtype: {self._dtype}>"
            )
        return repr(self._make_array())

    def __str__(self):
        # str alone should not force loading of the data
        if self._array is None:
            return (
                f"<{'array' if self._mask is None else 'masked array'} "
                f"(unloaded) shape: {self._shape} dtype: {self._dtype}>"
            )
        return str(self._make_array())

    def get_actual_shape(self, shape, strides, dtype, block_size):
        """
        Get the actual shape of an array, by computing it against the
        block_size if it contains a ``*``.
        """
        num_stars = shape.count("*")
        if num_stars == 0:
            return shape

        if num_stars == 1:
            if shape[0] != "*":
                msg = "'*' may only be in first entry of shape"
                raise ValueError(msg)

            stride = strides[0] if strides is not None else np.prod(shape[1:]) * dtype.itemsize

            missing = int(block_size / stride)
            return [missing] + shape[1:]

        msg = f"Invalid shape '{shape}'"
        raise ValueError(msg)

    @property
    def shape(self):
        if self._shape is None or self._array is not None or "*" in self._shape:
            # streamed blocks have a '0' data_size in the header so we
            # need to make the array to get the shape
            return self.__array__().shape
        return tuple(self._shape)

    @property
    def dtype(self):
        if self._array is None:
            return self._dtype

        return self._make_array().dtype

    def __len__(self):
        if self._array is None:
            return self._shape[0]

        return len(self._make_array())

    def __getattr__(self, attr):
        # We need to ignore __array_struct__, or unicode arrays end up
        # getting "double casted" and upsized.  This also reduces the
        # number of array creations in the general case.
        if attr == "__array_struct__":
            raise AttributeError
        # AsdfFile.info will call hasattr(obj, "__asdf_traverse__") which
        # will trigger this method, making the array, and loading the array
        # data. Intercept this and raise AttributeError as this class does
        # not support that method
        # see: https://github.com/asdf-format/asdf/issues/1553
        if attr == "__asdf_traverse__":
            raise AttributeError
        return getattr(self._make_array(), attr)

    def __setitem__(self, *args):
        # This workaround appears to be necessary in order to avoid a segfault
        # in the case that array assignment causes an exception. The segfault
        # originates from the call to __repr__ inside the traceback report.
        try:
            self._make_array().__setitem__(*args)
        except Exception:
            self._array = None

            raise


def _make_operation(name):
    def operation(self, *args):
        return getattr(self._make_array(), name)(*args)

    return operation


for op in [
    "__neg__",
    "__pos__",
    "__abs__",
    "__invert__",
    "__complex__",
    "__int__",
    "__long__",
    "__float__",
    "__oct__",
    "__hex__",
    "__lt__",
    "__le__",
    "__eq__",
    "__ne__",
    "__gt__",
    "__ge__",
    "__cmp__",
    "__rcmp__",
    "__add__",
    "__sub__",
    "__mul__",
    "__floordiv__",
    "__mod__",
    "__divmod__",
    "__pow__",
    "__lshift__",
    "__rshift__",
    "__and__",
    "__xor__",
    "__or__",
    "__div__",
    "__truediv__",
    "__radd__",
    "__rsub__",
    "__rmul__",
    "__rdiv__",
    "__rtruediv__",
    "__rfloordiv__",
    "__rmod__",
    "__rdivmod__",
    "__rpow__",
    "__rlshift__",
    "__rrshift__",
    "__rand__",
    "__rxor__",
    "__ror__",
    "__iadd__",
    "__isub__",
    "__imul__",
    "__idiv__",
    "__itruediv__",
    "__ifloordiv__",
    "__imod__",
    "__ipow__",
    "__ilshift__",
    "__irshift__",
    "__iand__",
    "__ixor__",
    "__ior__",
    "__getitem__",
    "__delitem__",
    "__contains__",
]:
    setattr(NDArrayType, op, _make_operation(op))


def _get_ndim(instance):
    if isinstance(instance, list):
        array = inline_data_asarray(instance)
        return array.ndim

    if isinstance(instance, dict):
        if "shape" in instance:
            return len(instance["shape"])

        if "data" in instance:
            array = inline_data_asarray(instance["data"])
            return array.ndim

    if isinstance(instance, (np.ndarray, NDArrayType)):
        return len(instance.shape)

    return None


def validate_ndim(validator, ndim, instance, schema):
    in_ndim = _get_ndim(instance)

    if in_ndim != ndim:
        yield ValidationError(f"Wrong number of dimensions: Expected {ndim}, got {in_ndim}", instance=repr(instance))


def validate_max_ndim(validator, max_ndim, instance, schema):
    in_ndim = _get_ndim(instance)

    if in_ndim is None or in_ndim > max_ndim:
        yield ValidationError(
            f"Wrong number of dimensions: Expected max of {max_ndim}, got {in_ndim}",
            instance=repr(instance),
        )


def validate_datatype(validator, datatype, instance, schema):
    if isinstance(instance, list):
        array = inline_data_asarray(instance)
        in_datatype, _ = numpy_dtype_to_asdf_datatype(array.dtype)
    elif isinstance(instance, dict):
        if "datatype" in instance:
            in_datatype = instance["datatype"]
        elif "data" in instance:
            array = inline_data_asarray(instance["data"])
            in_datatype, _ = numpy_dtype_to_asdf_datatype(array.dtype)
        else:
            msg = "Not an array"
            yield ValidationError(msg)
            return
    elif isinstance(instance, (np.ndarray, NDArrayType)):
        in_datatype, _ = numpy_dtype_to_asdf_datatype(instance.dtype)
    else:
        msg = "Not an array"
        yield ValidationError(msg)
        return

    # We are only concerned with some fields from the datatype
    # object in the schema so if the schema datatype is structured
    # copy the datatype and drop the irrelevant fields
    # name datatype byteorder shape
    if isinstance(datatype, list) and len(datatype) and isinstance(datatype[0], dict):
        datatype = [{k: v for k, v in subitem.items() if k in _STRUCTURED_DATATYPE_KEYS} for subitem in datatype]

    if datatype == in_datatype:
        return

    if schema.get("exact_datatype", False):
        yield ValidationError(f"Expected datatype '{datatype}', got '{in_datatype}'")
        return

    np_datatype = asdf_datatype_to_numpy_dtype(datatype)
    np_in_datatype = asdf_datatype_to_numpy_dtype(in_datatype)

    if not np_datatype.fields:
        if np_in_datatype.fields:
            yield ValidationError(f"Expected scalar datatype '{datatype}', got '{in_datatype}'")
            return

        if not np.can_cast(np_in_datatype, np_datatype, "safe"):
            yield ValidationError(f"Can not safely cast from '{in_datatype}' to '{datatype}' ")
            return

    else:
        if not np_in_datatype.fields:
            yield ValidationError(f"Expected structured datatype '{datatype}', got '{in_datatype}'")
            return

        if len(np_in_datatype.fields) != len(np_datatype.fields):
            yield ValidationError(f"Mismatch in number of columns: Expected {len(datatype)}, got {len(in_datatype)}")
            return

        for i in range(len(np_datatype.fields)):
            in_type = np_in_datatype[i]
            out_type = np_datatype[i]
            if not np.can_cast(in_type, out_type, "safe"):
                yield ValidationError(
                    "Can not safely cast to expected datatype: "
                    f"Expected {numpy_dtype_to_asdf_datatype(out_type)[0]}, "
                    f"got {numpy_dtype_to_asdf_datatype(in_type)[0]}",
                )
                return



"""
Utilities for displaying the content of an ASDF tree.

Normally these tools only will introspect dicts, lists, and primitive values
(with an exception for arrays). However, if the object that is generated
by the converter mechanism has a __asdf_traverse__() method, then it will
call that method expecting a dict or list to be returned. The method can
return what it thinks is suitable for display.
"""

import sys

from ._node_info import create_tree

__all__ = [
    "DEFAULT_MAX_COLS",
    "DEFAULT_MAX_ROWS",
    "DEFAULT_SHOW_VALUES",
    "render_tree",
]


DEFAULT_MAX_ROWS = 24
DEFAULT_MAX_COLS = 120
DEFAULT_SHOW_VALUES = True


def render_tree(
    node,
    max_rows=DEFAULT_MAX_ROWS,
    max_cols=DEFAULT_MAX_COLS,
    show_values=DEFAULT_SHOW_VALUES,
    filters=None,
    identifier="root",
    extension_manager=None,
):
    """
    Render a tree as text with indents showing depth.
    """
    info = create_tree(
        key="title",
        node=node,
        identifier=identifier,
        filters=[] if filters is None else filters,
        extension_manager=extension_manager,
    )
    if info is None:
        return []

    renderer = _TreeRenderer(
        max_rows,
        max_cols,
        show_values,
    )
    return renderer.render(info)


class _TreeRenderer:
    """
    Render a _NodeInfo tree with indent showing depth.
    """

    def __init__(self, max_rows, max_cols, show_values):
        self._max_rows = max_rows
        self._max_cols = max_cols
        self._show_values = show_values
        self._isatty = hasattr(sys.stdout, "isatty") and sys.stdout.isatty()

    def format_bold(self, value):
        """
        Wrap the input value in the ANSI escape sequence for increased intensity.
        """
        return self._format_code(value, 1)

    def format_faint(self, value):
        """
        Wrap the input value in the ANSI escape sequence for decreased intensity.
        """
        return self._format_code(value, 2)

    def format_italic(self, value):
        """
        Wrap the input value in the ANSI escape sequence for italic.
        """
        return self._format_code(value, 3)

    def _format_code(self, value, code):
        if not self._isatty:
            return f"{value}"
        return f"\x1b[{code}m{value}\x1b[0m"

    def render(self, info):
        self._mark_visible(info)

        lines, elided = self._render(info, set(), True)

        if elided:
            lines.append(self.format_faint(self.format_italic("Some nodes not shown.")))

        return lines

    def _mark_visible(self, root_info):
        """
        Select nodes to display, respecting max_rows.  Nodes at lower
        depths will be prioritized.
        """
        if isinstance(self._max_rows, tuple):
            self._mark_visible_tuple(root_info)
        else:
            self._mark_visible_int(root_info)

    def _mark_visible_int(self, root_info):
        """
        Select nodes to display, obeying max_rows as an overall limit on
        the number of lines returned.
        """
        if self._max_rows is None:
            return

        if self._max_rows < 2:
            root_info.visible = False
            return

        current_infos = [root_info]
        # Reserve one row for the root node, and another for the
        # "Some nodes not shown." message.
        rows_left = self._max_rows - 2
        while True:
            next_infos = []

            for info in current_infos:
                if rows_left >= len(info.children):
                    rows_left -= len(info.children)
                    next_infos.extend(info.children)
                elif rows_left > 1:
                    for child in info.children[rows_left - 1 :]:
                        child.visible = False
                    next_infos.extend(info.children[0 : rows_left - 1])
                    rows_left = 0
                else:
                    for child in info.children:
                        child.visible = False

            if len(next_infos) == 0:
                break

            current_infos = next_infos

    def _mark_visible_tuple(self, root_info):
        """
        Select nodes to display, obeying the per-node max_rows value for
        each tree depth.
        """
        max_rows = (None, *self._max_rows)

        current_infos = [root_info]
        while True:
            next_infos = []

            for info in current_infos:
                if info.depth + 1 < len(max_rows):
                    rows_left = max_rows[info.depth + 1]
                    if rows_left is None or rows_left >= len(info.children):
                        next_infos.extend(info.children)
                    elif rows_left > 1:
                        for child in info.children[rows_left - 1 :]:
                            child.visible = False
                        next_infos.extend(info.children[0 : rows_left - 1])
                    else:
                        for child in info.children:
                            child.visible = False
                else:
                    for child in info.children:
                        child.visible = False

            if len(next_infos) == 0:
                break

            current_infos = next_infos

    def _render(self, info, active_depths, is_tail):
        """
        Render the tree.  Called recursively on child nodes.

        is_tail indicates if the child is the last of the children,
        needed to indicate the proper connecting character in the tree
        display. Likewise, active_depths is used to track which preceding
        depths are incomplete thus need continuing lines preceding in
        the tree display.
        """
        lines = []

        if info.visible is False:
            return lines, True

        lines.append(self._render_node(info, active_depths, is_tail))

        elided = len(info.visible_children) < len(info.children)

        for i, child in enumerate(info.visible_children):
            if i == len(info.children) - 1:
                child_is_tail = True
                child_active_depths = active_depths
            else:
                child_is_tail = False
                child_active_depths = active_depths.union({info.depth})

            child_list, child_elided = self._render(child, child_active_depths, child_is_tail)
            lines.extend(child_list)
            elided = elided or child_elided

        num_visible_children = len(info.visible_children)
        if num_visible_children > 0 and num_visible_children != len(info.children):
            hidden_count = len(info.children) - num_visible_children
            prefix = self._make_prefix(info.depth + 1, active_depths, True)
            message = self.format_faint(self.format_italic(str(hidden_count) + " not shown"))
            lines.append(f"{prefix}{message}")

        return lines, elided

    def _render_node(self, info, active_depths, is_tail):
        prefix = self._make_prefix(info.depth, active_depths, is_tail)
        value = self._render_node_value(info)

        line = (
            f"{prefix}[{self.format_bold(info.identifier)}] {value}"
            if isinstance(info.parent_node, (list, tuple))
            else f"{prefix}{self.format_bold(info.identifier)} {value}"
        )

        if info.info is not None:
            line = line + self.format_faint(self.format_italic(" # " + info.info))
        visible_children = info.visible_children
        if len(visible_children) == 0 and len(info.children) > 0:
            line = line + self.format_italic(" ...")

        if info.recursive:
            line = line + " " + self.format_faint(self.format_italic("(recursive reference)"))

        if self._max_cols is not None and len(line) > self._max_cols:
            message = " (truncated)"
            line = line[0 : (self._max_cols - len(message))] + self.format_faint(self.format_italic(message))

        return line

    def _render_node_value(self, info):
        rendered_type = type(info.node).__name__

        if not info.children and self._show_values:
            try:
                s = f"{info.node}"
            except Exception:
                # if __str__ fails, don't fail info, instead use an empty string
                s = ""
            # if __str__ returns multiple lines also use an empty string
            if len(s.splitlines()) > 1:
                s = ""
            # if s is empty use the non-_show_values format below
            if s:
                return f"({rendered_type}): {s}"

        return f"({rendered_type})"

    def _make_prefix(self, depth, active_depths, is_tail):
        """
        Create a prefix for a displayed node, accounting for depth
        and including lines that show connections to other nodes.
        """
        prefix = ""

        if depth < 1:
            return prefix

        if depth >= 2:
            for n in range(0, depth - 1):
                prefix = prefix + "│ " if n in active_depths else prefix + "  "

        prefix = prefix + "└─" if is_tail else prefix + "├─"

        return self.format_faint(prefix)



import numpy as np

import asdf


def test_unable_to_read_empty_inline_array(tmp_path):
    """
    ASDF unable to read empty inline array

    https://github.com/asdf-format/asdf/issues/1538
    """
    fn = tmp_path / "test.asdf"
    a = np.array([])
    af = asdf.AsdfFile({"a": a})
    af.set_array_storage(a, "inline")
    af.write_to(fn)
    with asdf.open(fn) as af:
        np.testing.assert_array_equal(af["a"], a)



"""
Contains commands for lightweight text editing of an ASDF file.
"""

import io
import os
import re
import shutil

# Marked safe because the editor command is specified by an
# environment variable that the user controls.
import subprocess  # nosec
import sys
import tempfile

import yaml

from asdf import _io, constants, generic_io, schema, util
from asdf._asdf import AsdfFile
from asdf._block import io as bio
from asdf._block.exceptions import BlockIndexError

from .main import Command

__all__ = ["edit"]


if sys.platform.startswith("win"):
    DEFAULT_EDITOR = "notepad"
else:
    DEFAULT_EDITOR = "vi"


class Edit(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        """
        Set up a command line argument parser for the edit subcommand.
        """
        # Set up the parser
        parser = subparsers.add_parser(
            "edit",
            description="Edit the YAML portion of an ASDF file in-place.",
        )

        # Need an input file
        parser.add_argument(
            "filename",
            help="Path to an ASDF file.",
        )

        parser.set_defaults(func=cls.run)

        return parser

    @classmethod
    def run(cls, args):
        """
        Execute the edit subcommand.
        """
        return edit(args.filename)


def read_yaml(fd):
    """
    Read the YAML portion of an open ASDF file's content.

    Parameters
    ----------
    fd : GenericFile

    Returns
    -------
    bytes
        YAML content
    int
        total number of bytes available for YAML area
    bool
        True if the file contains binary blocks
    """
    # All ASDF files produced by this library, even the binary files
    # of an exploded ASDF file, include a YAML header, so we'll just
    # let this raise an error if the end marker can't be found.
    # Revisit this if someone starts producing files without a
    # YAML section, which the specification permits but is not possible
    # with current software.
    reader = fd.reader_until(
        constants.YAML_END_MARKER_REGEX,
        7,
        "End of YAML marker",
        include=True,
    )
    content = reader.read()

    reader = fd.reader_until(
        constants.BLOCK_MAGIC,
        len(constants.BLOCK_MAGIC),
        include=False,
        exception=False,
    )
    buffer = reader.read()

    contains_blocks = fd.peek(len(constants.BLOCK_MAGIC)) == constants.BLOCK_MAGIC

    return content, len(content) + len(buffer), contains_blocks


def write_edited_yaml_larger(path, new_content, version):
    """
    Rewrite an ASDF file, replacing the YAML portion with the
    specified YAML content and updating the block index if present.
    The file is assumed to contain binary blocks.

    Parameters
    ----------
    path : str
        Path to ASDF file
    content : bytes
        Updated YAML content
    """
    prefix = os.path.splitext(os.path.basename(path))[0] + "-"
    # Since the original file may be large, create the temporary
    # file in the same directory to avoid filling up the system
    # temporary area.
    temp_file = tempfile.NamedTemporaryFile(dir=os.path.dirname(path), prefix=prefix, suffix=".asdf", delete=False)
    try:
        temp_file.close()
        with generic_io.get_file(temp_file.name, mode="w") as fd:
            fd.write(new_content)
            # Allocate additional space for future YAML updates:
            pad_length = util.calculate_padding(len(new_content), True, fd.block_size)
            fd.fast_forward(pad_length)

            # now copy over ASDF block contents

            with generic_io.get_file(path) as original_fd:
                original_fd.seek_until(constants.BLOCK_MAGIC, len(constants.BLOCK_MAGIC))
                old_first_block_offset = original_fd.tell() - len(constants.BLOCK_MAGIC)
                new_first_block_offset = fd.tell()

                # check if the original file has a block index which we will need to update
                # as we're moving the blocks
                block_index_offset = bio.find_block_index(original_fd)
                if block_index_offset is None:
                    block_index = None
                    original_fd.seek(0, generic_io.SEEK_END)
                    blocks_end = original_fd.tell()
                else:
                    blocks_end = block_index_offset
                    try:
                        block_index = bio.read_block_index(original_fd, block_index_offset)
                    except BlockIndexError:
                        # the original index was invalid
                        block_index = None

                # copy over blocks byte-for-byte from old_first_block_offset to block_index_offset
                original_fd.seek(old_first_block_offset)
                block_size = min(fd.block_size, original_fd.block_size)
                n_bytes = blocks_end - old_first_block_offset
                for offset in range(0, n_bytes, block_size):
                    this_size = min(block_size, n_bytes - offset)
                    fd.write(original_fd.read(this_size))

                # update index
                if block_index is not None:
                    offset = new_first_block_offset - old_first_block_offset
                    updated_block_index = [i + offset for i in block_index]
                    bio.write_block_index(fd, updated_block_index)

        # Swap in the new version of the file atomically:
        shutil.copy(temp_file.name, path)
    finally:
        os.unlink(temp_file.name)


def write_edited_yaml(path, new_content, available_bytes):
    """
    Overwrite the YAML portion of an ASDF tree with the specified
    YAML content.  The content must fit in the space available.

    Parameters
    ----------
    path : str
        Path to ASDF file
    yaml_content : bytes
        Updated YAML content
    available_bytes : int
        Number of bytes available for YAML
    """
    # generic_io mode "rw" opens the file as "r+b":
    with generic_io.get_file(path, mode="rw") as fd:
        fd.write(new_content)

        pad_length = available_bytes - len(new_content)
        if pad_length > 0:
            fd.write(b"\0" * pad_length)


def edit(path):
    """
    Copy the YAML portion of an ASDF file to a temporary file, present
    the file to the user for editing, then update the original file
    with the modified YAML.

    Parameters
    ----------
    path : str
        Path to ASDF file
    """
    # Extract the YAML portion of the original file:
    with generic_io.get_file(path, mode="r") as fd:
        if fd.peek(len(constants.ASDF_MAGIC)) != constants.ASDF_MAGIC:
            print(f"Error: '{path}' is not an ASDF file.")
            return 1

        original_content, available_bytes, contains_blocks = read_yaml(fd)

    original_asdf_version = parse_asdf_version(original_content)
    original_yaml_version = parse_yaml_version(original_content)

    prefix = os.path.splitext(os.path.basename(path))[0] + "-"
    # We can't use temp_file's automatic delete because Windows
    # won't allow reading the file from the editor process unless
    # it is closed here.
    temp_file = tempfile.NamedTemporaryFile(prefix=prefix, suffix=".yaml", delete=False)
    try:
        # Write the YAML to a temporary path:
        temp_file.write(original_content)
        temp_file.close()

        # Loop so that the user can correct errors in the edited file:
        while True:
            open_editor(temp_file.name)

            with open(temp_file.name, "rb") as f:
                new_content = f.read()

            if new_content == original_content:
                print("No changes made to file")
                return 0

            try:
                new_asdf_version = parse_asdf_version(new_content)
                new_yaml_version = parse_yaml_version(new_content)
            except Exception as e:
                print(f"Error: failed to parse ASDF header: {e!s}")
                choice = request_input("(c)ontinue editing or (a)bort? ", ["c", "a"])
                if choice == "a":
                    return 1

                continue

            if new_asdf_version != original_asdf_version or new_yaml_version != original_yaml_version:
                print("Error: cannot modify ASDF Standard or YAML version using this tool.")
                choice = request_input("(c)ontinue editing or (a)bort? ", ["c", "a"])
                if choice == "a":
                    return 1

                continue

            try:
                # check this is an ASDF file
                if new_content[: len(constants.ASDF_MAGIC)] != constants.ASDF_MAGIC:
                    msg = "Does not appear to be a ASDF file."
                    raise ValueError(msg)
                # read the tagged tree (which also checks if the YAML is valid)
                tagged_tree = util.load_yaml(io.BytesIO(new_content), tagged=True)
                # validate the tagged tree
                ctx = AsdfFile(version=new_asdf_version)
                schema.validate(tagged_tree, ctx=ctx, reading=True)
            except yaml.YAMLError as e:
                print("Error: failed to parse updated YAML:")
                print_exception(e)
                choice = request_input("(c)ontinue editing or (a)bort? ", ["c", "a"])
                if choice == "a":
                    return 1

                continue

            except schema.ValidationError as e:
                print("Warning: updated ASDF tree failed validation:")
                print_exception(e)
                choice = request_input("(c)ontinue editing, (f)orce update, or (a)bort? ", ["c", "f", "a"])
                if choice == "a":
                    return 1

                if choice == "c":
                    continue

            except Exception as e:
                print("Error: failed to read updated file as ASDF:")
                print_exception(e)
                choice = request_input("(c)ontinue editing or (a)bort? ", ["c", "a"])
                if choice == "a":
                    return 1

                continue

            # We've either opened the file without error, or
            # the user has agreed to ignore validation errors.
            # Break out of the loop so that we can update the
            # original file.
            break
    finally:
        os.unlink(temp_file.name)

    if len(new_content) <= available_bytes:
        # File has sufficient space allocated in the YAML area.
        write_edited_yaml(path, new_content, available_bytes)
    elif not contains_blocks:
        # File does not have sufficient space, but there are
        # no binary blocks, so we can just expand the file.
        write_edited_yaml(path, new_content, len(new_content))
    else:
        # File does not have sufficient space, and binary blocks
        # are present.
        print("Warning: updated YAML larger than allocated space.  File must be rewritten.")
        choice = request_input("(c)ontinue or (a)bort? ", ["c", "a"])
        if choice == "a":
            return 1

        write_edited_yaml_larger(path, new_content, new_asdf_version)

    return None


def parse_asdf_version(content):
    """
    Extract the ASDF core schemas version from YAML content.

    Parameters
    ----------
    content : bytes

    Returns
    -------
    asdf.versioning.AsdfVersion
        ASDF core schemas version
    """
    comments = _io.read_comment_section(generic_io.get_file(io.BytesIO(content)))
    return _io.find_asdf_version_in_comments(comments)


def parse_yaml_version(content):
    """
    Extract the YAML version from YAML content.

    Parameters
    ----------
    content : bytes

    Returns
    -------
    bytes
        YAML version string.
    """
    match = re.search(b"^%YAML (.*)$", content, flags=re.MULTILINE)
    if match is None:
        msg = "YAML version number not found"
        raise ValueError(msg)
    return match.group(1)


def print_exception(e):
    """
    Print an exception, indented 4 spaces and elided if too many lines.
    """
    lines = str(e).split("\n")
    if len(lines) > 20:
        lines = lines[0:20] + ["..."]
    for line in lines:
        print(f"    {line}")


def request_input(message, choices):
    """
    Request user input.

    Parameters
    ----------
    message : str
        Message to display
    choices : list of str
        List of recognized inputs
    """
    while True:
        choice = input(message).strip().lower()

        if choice in choices:
            return choice

        print(f"Invalid choice: {choice}")

    return None


def open_editor(path):
    """
    Launch an editor process with the file at path opened.
    """
    editor = os.environ.get("EDITOR", DEFAULT_EDITOR)
    # Marked safe because the editor command is specified by an
    # environment variable that the user controls.
    subprocess.run(f"{editor} {path}", check=True, shell=True)  # noqa: S602



import contextlib
import copy
import getpass
import io
import os
import pathlib
import sys

import numpy as np
import pytest
from numpy.testing import assert_array_equal

import asdf
from asdf import config_context, get_config, treeutil, versioning
from asdf.exceptions import AsdfPackageVersionWarning, ValidationError
from asdf.extension import ExtensionProxy
from asdf.resource import ResourceMappingProxy
from asdf.testing.helpers import roundtrip_object, yaml_to_asdf

from ._helpers import assert_tree_match

RNG = np.random.default_rng(97)


def test_get_data_from_closed_file(tmp_path):
    path = str(tmp_path / "test.asdf")

    my_array = np.arange(0, 64).reshape((8, 8))

    tree = {"my_array": my_array}
    ff = asdf.AsdfFile(tree)
    ff.write_to(path)

    with asdf.open(path) as ff:
        pass

    with pytest.raises(IOError, match=r"Cannot access data from closed ASDF file"):
        assert_array_equal(my_array, ff.tree["my_array"])


def test_no_warning_nan_array(tmp_path):
    """
    Tests for a regression that was introduced by
    https://github.com/asdf-format/asdf/pull/557
    where saving an array with a nan resulted in a warning
    """
    roundtrip_object(np.array([1, 2, np.nan]))


@pytest.mark.skipif(
    not sys.platform.startswith("win") and getpass.getuser() == "root",
    reason="Cannot make file read-only if user is root",
)
def test_open_readonly(tmp_path):
    tmpfile = str(tmp_path / "readonly.asdf")

    tree = {"foo": 42, "bar": "hello", "baz": np.arange(20)}
    with asdf.AsdfFile(tree) as af:
        af.write_to(tmpfile, all_array_storage="internal")

    os.chmod(tmpfile, 0o440)
    assert os.access(tmpfile, os.W_OK) is False

    with asdf.open(tmpfile, memmap=True) as af:
        assert af["baz"].flags.writeable is False

    with pytest.raises(PermissionError, match=r".* Permission denied: .*"), asdf.open(tmpfile, mode="rw"):
        pass


def test_open_validate_on_read(tmp_path):
    content = """
invalid_software: !core/software-1.0.0
  name: Minesweeper
  version: 3
"""
    buff = yaml_to_asdf(content)

    get_config().validate_on_read = True
    with pytest.raises(ValidationError, match=r".* is not of type .*"), asdf.open(buff):
        pass

    buff.seek(0)

    get_config().validate_on_read = False
    with asdf.open(buff) as af:
        assert af["invalid_software"]["name"] == "Minesweeper"
        assert af["invalid_software"]["version"] == 3


def test_open_stream(tmp_path):
    file_path = tmp_path / "test.asdf"

    with asdf.AsdfFile() as af:
        af["foo"] = "bar"
        af.write_to(file_path)

    class StreamWrapper:
        def __init__(self, fd):
            self._fd = fd

        def read(self, size=-1):
            return self._fd.read(size)

    with file_path.open("rb") as fd, asdf.open(StreamWrapper(fd)) as af:
        assert af["foo"] == "bar"


def test_atomic_write(tmp_path, small_tree):
    tmpfile = str(tmp_path / "test.asdf")

    ff = asdf.AsdfFile(small_tree)
    ff.write_to(tmpfile)

    with asdf.open(tmpfile, mode="r") as ff:
        ff.write_to(tmpfile)


def test_default_version():
    """
    See https://github.com/asdf-format/asdf/issues/364
    """
    ff = asdf.AsdfFile()
    assert ff.file_format_version == versioning._FILE_FORMAT_VERSION

    # make sure these are different AsdfVersion instances
    assert ff.file_format_version is not versioning._FILE_FORMAT_VERSION


def test_update_exceptions(tmp_path):
    path = str(tmp_path / "test.asdf")

    my_array = RNG.normal(size=(8, 8))
    tree = {"my_array": my_array}
    ff = asdf.AsdfFile(tree)
    ff.write_to(path)

    with (
        asdf.open(path, mode="r", memmap=False) as ff,
        pytest.raises(
            IOError,
            match=r"Can not update, since associated file is read-only.*",
        ),
    ):
        ff.update()

    ff = asdf.AsdfFile(tree)
    buff = io.BytesIO()
    ff.write_to(buff)

    buff.seek(0)
    with asdf.open(buff, mode="rw") as ff:
        ff.update()

    with pytest.raises(ValueError, match=r"Can not update, since there is no associated file"):
        asdf.AsdfFile().update()


def test_top_level_tree(small_tree):
    tree = {"tree": small_tree}
    ff = asdf.AsdfFile(tree)
    assert_tree_match(ff.tree["tree"], ff["tree"])

    ff2 = asdf.AsdfFile()
    ff2["tree"] = small_tree
    assert_tree_match(ff2.tree["tree"], ff2["tree"])


def test_top_level_keys(small_tree):
    tree = {"tree": small_tree}
    ff = asdf.AsdfFile(tree)
    assert ff.tree.keys() == ff.keys()


def test_top_level_contains():
    tree = {
        "foo": 42,
        "bar": 43,
    }

    with asdf.AsdfFile(tree) as af:
        assert "foo" in af
        assert "bar" in af


def test_walk_and_modify_remove_keys():
    tree = {"foo": 42, "bar": 43}

    def func(x):
        if x == 42:
            return treeutil.RemoveNode
        return x

    tree2 = treeutil.walk_and_modify(tree, func)

    assert "foo" not in tree2
    assert "bar" in tree2


def test_walk_and_modify_retain_none():
    tree = {"foo": 42, "bar": None}

    def func(x):
        if x == 42:
            return None
        return x

    tree2 = treeutil.walk_and_modify(tree, func)

    assert tree2["foo"] is None
    assert tree2["bar"] is None


def test_copy(tmp_path):
    my_array = RNG.normal(size=(8, 8))
    tree = {"my_array": my_array, "foo": {"bar": "baz"}}
    ff = asdf.AsdfFile(tree)
    ff.write_to(str(tmp_path / "test.asdf"))

    with asdf.open(str(tmp_path / "test.asdf")) as ff:
        ff2 = ff.copy()
        ff2.tree["my_array"] *= 2
        ff2.tree["foo"]["bar"] = "boo"

        assert np.all(ff2.tree["my_array"] == ff.tree["my_array"] * 2)
        assert ff.tree["foo"]["bar"] == "baz"

    assert_array_equal(ff2.tree["my_array"], ff2.tree["my_array"])


def test_access_tree_outside_handler(tmp_path):
    tempname = str(tmp_path / "test.asdf")

    tree = {"random": np.random.random(10)}

    ff = asdf.AsdfFile(tree)
    ff.write_to(str(tempname))

    with asdf.open(tempname) as newf:
        pass

    # Accessing array data outside of handler should fail
    with pytest.raises(OSError, match=r"Cannot access data from closed ASDF file"):
        repr(newf.tree["random"])

    # Using the top-level getattr should also fail
    with pytest.raises(OSError, match=r"Cannot access data from closed ASDF file"):
        repr(newf["random"])


def test_context_handler_resolve_and_inline(tmp_path):
    """
    This reproduces the issue reported in
    https://github.com/asdf-format/asdf/issues/406
    """
    tempname = str(tmp_path / "test.asdf")

    tree = {"random": np.random.random(10)}

    ff = asdf.AsdfFile(tree)
    ff.write_to(str(tempname))

    with asdf.open(tempname) as newf:
        newf.resolve_references()
        newf.set_array_storage(newf["random"], "inline")

    with pytest.raises(OSError, match=r"Cannot access data from closed ASDF file"):
        newf.tree["random"][0]


def test_open_pathlib_path(tmp_path):
    filename = str(tmp_path / "pathlib.asdf")
    path = pathlib.Path(filename)

    tree = {"data": np.ones(10)}

    with asdf.AsdfFile(tree) as af:
        af.write_to(path)

    with asdf.open(path) as af:
        assert (af["data"] == tree["data"]).all()


@pytest.mark.parametrize("strict", [True, False])
@pytest.mark.parametrize(
    ("installed", "extension", "warns"),
    [
        (None, "2.0.0", True),
        ("1.2.3", "2.0.0", True),
        ("1.2.3", "2.0.dev10842", True),
        ("2.0.0", "2.0.0", False),
        ("2.0.1", "2.0.0", False),
        ("2.0.1", "2.0.dev12345", False),
    ],
)
def test_extension_version_check(installed, extension, warns, strict, with_lazy_tree, tmp_path):
    class FooExtension:
        extension_uri = "asdf://somewhere.org/extensions/foo-1.0.0"

    proxy = ExtensionProxy(FooExtension(), package_name="foo", package_version=installed)

    test_filename = tmp_path / "test.asdf"

    asdf.dump(
        {
            "history": {
                "extensions": [
                    asdf.tags.core.ExtensionMetadata(
                        extension_class="something",
                        extension_uri=FooExtension.extension_uri,
                        software=asdf.tags.core.Software(name="foo", version=extension),
                    ),
                ],
            },
        },
        test_filename,
    )
    if warns:
        if strict:
            ctx = pytest.raises(RuntimeError, match=r"was created with extension")
        else:
            ctx = pytest.warns(AsdfPackageVersionWarning, match=r"was created with extension")
    else:
        ctx = contextlib.nullcontext()

    with config_context() as config:
        if installed is not None:
            config.add_extension(proxy)
        with ctx, asdf.open(test_filename, strict_extension_check=strict):
            pass


@pytest.mark.parametrize(
    ("installed", "extension", "warns"),
    [
        (None, "2.0.0", True),
        ("1.2.3", "2.0.0", True),
        ("1.2.3", "2.0.dev10842", True),
        ("2.0.0", "2.0.0", False),
        ("2.0.1", "2.0.0", False),
        ("2.0.1", "2.0.dev12345", False),
    ],
)
def test_check_extension_manifest_software(installed, extension, warns):
    class FooExtension:
        extension_uri = "asdf://somewhere.org/extensions/foo-1.0.0"

    proxy = ExtensionProxy(FooExtension(), package_name="foo", package_version="1.0.0")

    mapping = ResourceMappingProxy({}, package_name="bar", package_version=installed)

    with config_context() as config:
        config.add_extension(proxy)
        if installed is not None:
            config.add_resource_mapping(mapping)
        af = asdf.AsdfFile()

        tree = {
            "history": {
                "extensions": [
                    asdf.tags.core.ExtensionMetadata(
                        extension_uri=FooExtension.extension_uri,
                        software=asdf.tags.core.Software(name="foo", version="1.0.0"),
                        manifest_software=asdf.tags.core.Software(name="bar", version=extension),
                    ),
                ],
            },
        }

        if warns:
            with pytest.warns(AsdfPackageVersionWarning, match=r"File was created with"):
                af._check_extensions(tree)

            with pytest.raises(RuntimeError, match=r"^File was created with"):
                af._check_extensions(tree, strict=True)

        else:
            af._check_extensions(tree)


def test_extension_check_no_warning_on_builtin():
    """
    Prior to asdf 3.0 files were written using the asdf.extension.BuiltinExtension
    (which used the legacy extension api). This extension was removed in
    asdf 3.0. We don't want to warn that this extension is missing for every
    file that is opened so make sure _check_extensions doesn't warn
    that BuiltinExtension is missing.
    """
    af = asdf.AsdfFile()
    tree = {
        "history": {
            "extensions": [
                asdf.tags.core.ExtensionMetadata(
                    extension_class="asdf.extension.BuiltinExtension",
                    software=asdf.tags.core.Software(name="asdf", version="2.15.1"),
                ),
            ],
        },
    }
    af._check_extensions(tree)


@pytest.mark.parametrize(
    ("array_inline_threshold", "inline_blocks", "internal_blocks"),
    [
        (None, 0, 2),
        (10, 1, 1),
        (7, 1, 1),
        (5, 0, 2),
        (0, 0, 2),
        (1, 0, 2),
    ],
)
def test_array_inline_threshold(array_inline_threshold, inline_blocks, internal_blocks, tmp_path):
    file_path = tmp_path / "test.asdf"
    tree = {"small_array": np.arange(6), "large_array": np.arange(100)}

    with asdf.config_context() as config:
        config.array_inline_threshold = array_inline_threshold

        with asdf.AsdfFile(tree) as af:
            af.write_to(file_path)
        with asdf.open(file_path) as af:
            assert len(af._blocks.blocks) == internal_blocks


@pytest.mark.parametrize(
    ("array_inline_threshold", "inline_blocks", "internal_blocks"),
    [
        (None, 0, 2),
        (10, 2, 0),
        (5, 0, 2),
    ],
)
def test_array_inline_threshold_masked_array(array_inline_threshold, inline_blocks, internal_blocks, tmp_path):
    file_path = tmp_path / "test.asdf"
    arr = np.arange(6)
    masked_arr = np.ma.masked_equal(arr, 3)
    tree = {"masked_arr": masked_arr}

    with asdf.config_context() as config:
        config.array_inline_threshold = array_inline_threshold

        with asdf.AsdfFile(tree) as af:
            af.write_to(file_path)
        with asdf.open(file_path) as af:
            assert len(af._blocks.blocks) == internal_blocks


@pytest.mark.parametrize(
    ("array_inline_threshold", "inline_blocks", "internal_blocks"),
    [
        (None, 0, 1),
        (10, 1, 0),
        (5, 0, 1),
    ],
)
def test_array_inline_threshold_string_array(array_inline_threshold, inline_blocks, internal_blocks, tmp_path):
    file_path = tmp_path / "test.asdf"
    arr = np.array(["peach", "plum", "apricot", "nectarine", "cherry", "pluot"])
    tree = {"array": arr}

    with asdf.config_context() as config:
        config.array_inline_threshold = array_inline_threshold

        with asdf.AsdfFile(tree) as af:
            af.write_to(file_path)
        with asdf.open(file_path) as af:
            assert len(af._blocks.blocks) == internal_blocks


def test_history_entries(tmp_path):
    path = str(tmp_path / "test.asdf")
    message = "Twas brillig, and the slithy toves"

    af = asdf.AsdfFile()
    af.add_history_entry(message)
    af.write_to(path)
    with asdf.open(path) as af:
        assert af["history"]["entries"][0]["description"] == message

    af = asdf.AsdfFile()
    af.write_to(path)
    with asdf.open(path) as af:
        af.add_history_entry(message)
        assert af["history"]["entries"][0]["description"] == message


def test_array_access_after_file_close(tmp_path):
    path = str(tmp_path / "test.asdf")
    data = np.arange(10)
    asdf.AsdfFile({"data": data}).write_to(path)

    # Normally it's not possible to read the array after
    # the file has been closed:
    with asdf.open(path) as af:
        tree = af.tree
    with pytest.raises(OSError, match=r"ASDF file has already been closed. Can not get the data."):
        tree["data"][0]

    # With memory mapping disabled and copying arrays enabled,
    # the array data should still persist in memory after close:
    with asdf.open(path, lazy_load=False, memmap=False) as af:
        tree = af.tree
    assert_array_equal(tree["data"], data)


def test_none_values(tmp_path):
    path = str(tmp_path / "test.asdf")

    af = asdf.AsdfFile({"foo": None})
    af.write_to(path)
    with asdf.open(path) as af:
        assert "foo" in af
        assert af["foo"] is None


def test_asdf_standard_version_tag_selection():
    buff = io.BytesIO()

    af = asdf.AsdfFile()
    af.write_to(buff, version="1.0.0")
    buff.seek(0)
    content = buff.read()
    assert b"!core/asdf-1.0.0" in content
    assert b"!core/asdf-1.1.0" not in content

    buff.seek(0)
    af.write_to(buff, version="1.2.0")  # asdf-standard 1.2 uses asdf-object 1.1 tag
    buff.seek(0)
    content = buff.read()
    assert b"!core/asdf-1.0.0" not in content
    assert b"!core/asdf-1.1.0" in content


def test_update_asdf_standard_version_tag_selection():
    buff = io.BytesIO()

    af = asdf.AsdfFile()
    af.write_to(buff, version="1.0.0")

    buff.seek(0)
    with asdf.open(buff, mode="rw") as af:
        af.update(version="1.2.0")  # asdf-standard 1.2 uses asdf-object 1.1 tag
    buff.seek(0)
    content = buff.read()
    assert b"!core/asdf-1.1.0" in content
    assert b"!core/asdf-1.0.0" not in content


@pytest.mark.parametrize("valid_filename", [True, False], ids=["valid_filename", "invalid_filename"])
def test_write_to_no_tree_modification(tmp_path, valid_filename):
    if valid_filename:
        fn = tmp_path / "test.asdf"
    else:
        fn = "invalid/missing.asdf"
    fn2 = tmp_path / "test2.asdf"
    tree = {"foo": None}
    af = asdf.AsdfFile(tree.copy())
    try:
        af.write_to(fn)
    except Exception:
        if valid_filename:
            raise
    assert tree == af.tree
    if not valid_filename:
        return
    with asdf.open(fn) as af:
        af["history"]["extensions"][0]["software"]["version"] = "0.0.0.dev+abcdefg"
        af["asdf_library"]["author"] = "foo"
        tree = copy.deepcopy(af.tree)
        af.write_to(fn2)
        assert af.tree == tree


@pytest.mark.parametrize("valid_filename", [True, False], ids=["valid_filename", "invalid_filename"])
def test_write_to_no_version_modification(tmp_path, valid_filename):
    if valid_filename:
        fn = tmp_path / "test.asdf"
    else:
        fn = "invalid/missing.asdf"
    tree = {"foo": None}
    af = asdf.AsdfFile(tree.copy(), version="1.0.0")
    try:
        af.write_to(fn, version="1.1.0")
    except Exception:
        if valid_filename:
            raise
    assert af.version_string == "1.0.0"
    if not valid_filename:
        return
    with asdf.open(fn) as af:
        assert af.version_string == "1.1.0"



import io

import pytest

from asdf import AsdfFile
from asdf._commands import list_tags


@pytest.mark.parametrize("display_classes", [True, False])
def test_parameter_combinations(display_classes):
    # Just confirming no errors:
    list_tags(display_classes)


def test_all_tags_present():
    iostream = io.StringIO()
    list_tags(iostream=iostream)
    iostream.seek(0)
    tags = {line.strip() for line in iostream.readlines()}

    af = AsdfFile()
    for tag in af.extension_manager._converters_by_tag:
        assert tag in tags



import pytest
import yaml

import asdf
from asdf._core._integration import get_extensions, get_json_schema_resource_mappings


@pytest.mark.parametrize(
    "uri",
    [
        "http://json-schema.org/draft-04/schema",
    ],
)
def test_get_resource_mappings(uri):
    mappings = get_json_schema_resource_mappings()

    mapping = next(m for m in mappings if uri in m)
    assert mapping is not None

    assert uri.encode("utf-8") in mapping[uri]


def test_get_extensions():
    extensions = get_extensions()
    extension_uris = {e.extension_uri for e in extensions}

    # No duplicates
    assert len(extension_uris) == len(extensions)

    resource_extension_uris = set()
    resource_manager = asdf.get_config().resource_manager
    for resource_uri in resource_manager:
        if resource_uri.startswith("asdf://asdf-format.org/core/manifests/core-"):
            resource_extension_uris.add(yaml.safe_load(resource_manager[resource_uri])["extension_uri"])

    # Make sure every core manifest has a corresponding extension
    assert resource_extension_uris == extension_uris



import pytest

from asdf._commands import main


def test_help():
    # Just a smoke test, really
    main.main_from_args(["help"])


def test_invalid_command():
    with pytest.raises(SystemExit) as e:
        main.main([])
    assert e.value.code == 2

    with pytest.raises(SystemExit) as e:
        main.main(["foo"])
    if isinstance(e.value, int):
        assert e.value == 2
    else:
        assert e.value.code == 2



import numpy as np

ASDF_MAGIC = b"#ASDF"
BLOCK_MAGIC = b"\xd3BLK"
FITS_MAGIC = b"SIMPLE"

BLOCK_HEADER_BOILERPLATE_SIZE = 6

ASDF_STANDARD_COMMENT = b"ASDF_STANDARD"

INDEX_HEADER = b"#ASDF BLOCK INDEX"

# The maximum number of blocks supported
MAX_BLOCKS = 2**16
MAX_BLOCKS_DIGITS = int(np.ceil(np.log10(MAX_BLOCKS) + 1))

YAML_TAG_PREFIX = "tag:yaml.org,2002:"
YAML_END_MARKER_REGEX = rb"\r?\n\.\.\.((\r?\n)|$)"


STSCI_SCHEMA_URI_BASE = "http://stsci.edu/schemas/"
STSCI_SCHEMA_TAG_BASE = "tag:stsci.edu:asdf"


BLOCK_FLAG_STREAMED = 0x1


# ASDF max number size
MAX_BITS = 63
MAX_NUMBER = (1 << MAX_BITS) - 1
MIN_NUMBER = -((1 << MAX_BITS) - 2)



import re
from collections import namedtuple

from .schema import load_schema
from .treeutil import get_children, is_container


def _filter_tree(info, filters):
    """
    Keep nodes in tree which pass the all of the filters.

    Mutates the tree.
    """
    filtered_children = []
    for child in info.children:
        if _filter_tree(child, filters):
            filtered_children.append(child)
    info.children = filtered_children

    return len(info.children) > 0 or all(f(info.node, info.identifier) for f in filters)


def _get_matching_schema_property(schema, property_name):
    """
    Extract a property subschema for a given property_name.
    This function does not descend into the schema (beyond
    looking for a "properties" key) and does not support
    schema combiners.

    Parameters
    ----------
    schema : dict
        A dictionary containing a JSONSCHEMA
    property_name : str
        The name of the property to extract

    Returns
    -------
    dict or None
        The property subschema at the provided name or
        ``None`` if the property doesn't exist.
    """
    if "properties" in schema:
        props = schema["properties"]
        if property_name in props:
            return props[property_name]
        if "patternProperties" in props:
            patterns = props["patternProperties"]
            for regex in patterns:
                if re.search(regex, property_name):
                    return patterns[regex]
    return None


def _get_subschema_for_property(schema, property_name):
    """
    Extract a property subschema for a given property_name.
    This function will attempt to consider schema combiners
    and will return None on an ambiguous result.

    Parameters
    ----------
    schema : dict
        A dictionary containing a JSONSCHEMA
    property_name : str
        The name of the property to extract

    Returns
    -------
    dict or None
        The property subschema at the provided name or
        ``None`` if the property doesn't exist or is
        ambiguous (has more than one subschema or is nested in a not).
    """
    # This does NOT handle $ref the expectation is that the schema
    # is loaded with resolve_references=True
    applicable = []

    # first check properties and patternProperties
    subschema = _get_matching_schema_property(schema, property_name)
    if subschema is not None:
        applicable.append(subschema)

    # next handle schema combiners
    if "not" in schema:
        subschema = _get_subschema_for_property(schema["not"], property_name)
        if subschema is not None:
            # We can't resolve a valid subschema under a "not" since
            # we'd have to know how to invert a schema
            return None

    for combiner in ("allOf", "oneOf", "anyOf"):
        for combined_schema in schema.get(combiner, []):
            subschema = _get_subschema_for_property(combined_schema, property_name)
            if subschema is not None:
                applicable.append(subschema)

    # only return the subschema if we found exactly 1 applicable
    if len(applicable) == 1:
        return applicable[0]
    return None


def _get_schema_key(schema, key):
    """
    Extract a subschema at a given key.
    This function will attempt to consider schema combiners
    (allOf, oneOf, anyOf) and will return None on an
    ambiguous result (where more than 1 match is found).

    Parameters
    ----------
    schema : dict
        A dictionary containing a JSONSCHEMA
    key : str
        The key under which the subschema is stored

    Returns
    -------
    dict or None
        The subschema at the provided key or
        ``None`` if the key doesn't exist or is ambiguous.
    """
    applicable = []
    if key in schema:
        applicable.append(schema[key])
    # Here we don't consider any subschema under "not" to avoid
    # false positives for keys like "type" etc.
    for combiner in ("allOf", "oneOf", "anyOf"):
        for combined_schema in schema.get(combiner, []):
            possible = _get_schema_key(combined_schema, key)
            if possible is not None:
                applicable.append(possible)

    # only return the property if we found exactly 1 applicable
    if len(applicable) == 1:
        return applicable[0]
    return None


def create_tree(key, node, identifier="root", filters=None, extension_manager=None):
    """
    Create a `NodeSchemaInfo` tree which can be filtered from a base node.

    Parameters
    ----------
    key : str
        The key to look up.
    node : dict
        The asdf tree to search.
    filters : list of functions
        A list of functions that take a node and identifier and return True if the node should be included in the tree.
    preserve_list : bool
        If True, then lists are preserved. Otherwise, they are turned into dicts.
    """
    filters = [] if filters is None else filters

    schema_info = NodeSchemaInfo.from_root_node(
        key,
        identifier,
        node,
        extension_manager=extension_manager,
    )

    if len(filters) > 0 and not _filter_tree(schema_info, filters):
        return None

    return schema_info


def collect_schema_info(
    key,
    path,
    node,
    identifier="root",
    filters=None,
    preserve_list=True,
    extension_manager=None,
):
    """
    Collect from the underlying schemas any of the info stored under key, relative to the path

    Parameters
    ----------
    key : str
        The key to look up.
    path : str or None
        A dot-separated path to the parameter to find the key information on.
        If None return full dictionary.
    node : dict
        The asdf tree to search.
    filters : list of functions
        A list of functions that take a node and identifier and return True if the node should be included in the tree.
    preserve_list : bool
        If True, then lists are preserved. Otherwise, they are turned into dicts.
    """

    schema_info = create_tree(
        key,
        node,
        identifier=identifier,
        filters=[] if filters is None else filters,
        extension_manager=extension_manager,
    )

    info = schema_info.collect_info(preserve_list=preserve_list)

    if path is not None:
        for path_key in path.split("."):
            info = info.get(path_key, None)

            if info is None:
                return None

    return info


def _get_extension_manager():
    from ._asdf import AsdfFile

    af = AsdfFile()
    return af.extension_manager


def _make_traversable(node, extension_manager):
    if hasattr(node, "__asdf_traverse__"):
        return node.__asdf_traverse__(), True, False
    node_type = type(node)
    if not extension_manager.handles_type(node_type):
        return node, False, False
    return extension_manager.get_converter_for_type(node_type).to_info(node), False, True


_SchemaInfo = namedtuple("SchemaInfo", ["info", "value"])


class SchemaInfo(_SchemaInfo):
    """
    A class to hold the schema info and the value of the node.

    Parameters
    ----------
    info : dict
        The schema info.
    value : object
        The value of the node.
    """

    def __repr__(self):
        return f"{self.info}"


class NodeSchemaInfo:
    """
    Container for keyed information collected from a schema about a node of an ASDF file tree.

        This contains node alongside the parent and child nodes of that node in the ASDF file tree.
        Effectively this means that each of these "node" objects represents the a subtree of the file tree
        rooted at the node in question alongside methods to access the underlying schemas for the portions
        of the ASDF file in question.

    This is used for a variety of general purposes, including:
    - Providing the long descriptions for nodes as described in the schema.
    - Assisting in traversing an ASDF file like trees, to search nodes.
    - Providing a way to pull static information about an ASDF file which has
      been stored within the schemas for that file.

    Parameters
    ----------
    key : str
        The key for the information to be collected from the underlying schema(s).

    parent : NodeSchemaInfo
        The parent node of this node. None if this is the root node.

    identifier : str
        The identifier for this node in the ASDF file tree.

    node : any
        The value of the node in the ASDF file tree.

    depth : int
        The depth of this node in the ASDF file tree.

    recursive : bool
        If this node has already been visited, then this is set to True. Default is False.

    visible : bool
        If this node will be made visible in the output. Default is True.

    children : list
        List of the NodeSchemaInfo objects for the children of this node. This is a leaf node if this is empty.

    schema : dict
        The portion of the underlying schema corresponding to the node.
    """

    def __init__(self, key, parent, identifier, node, depth, recursive=False, visible=True, extension_manager=None):
        self.key = key
        self.parent = parent
        self.identifier = identifier
        self.node = node
        self.depth = depth
        self.recursive = recursive
        self.visible = visible
        self.children = []
        self.schema = None
        self.extension_manager = extension_manager or _get_extension_manager()

    @property
    def visible_children(self):
        return [c for c in self.children if c.visible]

    @property
    def parent_node(self):
        if self.parent is not None:
            return self.parent.node

        return None

    @property
    def info(self):
        if self.schema is None:
            return None
        return _get_schema_key(self.schema, self.key)

    def get_schema_for_property(self, identifier):
        return _get_subschema_for_property(self.schema, identifier) or {}

    def set_schema_for_property(self, parent, identifier):
        """Extract a subschema from the parent for the identified property"""

        self.schema = parent.get_schema_for_property(identifier)

    def set_schema_from_node(self, node, extension_manager):
        """Pull a tagged schema for the node"""

        tag_def = extension_manager.get_tag_definition(node._tag)
        schema_uri = tag_def.schema_uris[0]
        schema = load_schema(schema_uri, resolve_references=True)

        self.schema = schema

    @classmethod
    def from_root_node(cls, key, root_identifier, root_node, schema=None, extension_manager=None):
        """
        Build a NodeSchemaInfo tree from the given ASDF root node.
        Intentionally processes the tree in breadth-first order so that recursively
        referenced nodes are displayed at their shallowest reference point.
        """
        extension_manager = extension_manager or _get_extension_manager()

        current_nodes = [(None, root_identifier, root_node)]
        seen = set()
        root_info = None
        current_depth = 0
        while True:
            next_nodes = []

            for parent, identifier, node in current_nodes:
                # node is the item in the tree
                # We might sometimes not want to use that node directly
                # but instead using a different node for traversal.
                t_node, traversable, _ = _make_traversable(node, extension_manager)
                if (is_container(node) or traversable) and id(node) in seen:
                    info = NodeSchemaInfo(
                        key,
                        parent,
                        identifier,
                        node,
                        current_depth,
                        recursive=True,
                        extension_manager=extension_manager,
                    )
                    parent.children.append(info)

                else:
                    info = NodeSchemaInfo(
                        key, parent, identifier, node, current_depth, extension_manager=extension_manager
                    )

                    # If this is the first node keep a reference so we can return it.
                    if root_info is None:
                        root_info = info

                    if parent is None:
                        info.schema = schema

                    if parent is not None:
                        if parent.schema is not None:
                            # descend within the schema of the parent
                            info.set_schema_for_property(parent, identifier)

                        # track that this node is a child of the parent
                        parent.children.append(info)

                    # Track which nodes have been seen to avoid an infinite
                    # loop and to find recursive references
                    # This is tree wide but should be per-branch.
                    seen.add(id(node))

                    # if the node has __asdf_traverse__ and a _tag attribute
                    # that is a valid tag, load it's schema
                    if traversable:
                        if hasattr(node, "_tag") and isinstance(node._tag, str):
                            try:
                                info.set_schema_from_node(node, extension_manager)
                            except KeyError:
                                # if _tag is not a valid tag, no schema will be found
                                # and a KeyError will be raised. We don't raise the KeyError
                                # (or another exception) here as the node (object) might
                                # be using _tag for a non-ASDF purpose.
                                pass

                    # add children to queue
                    for child_identifier, child_node in get_children(t_node):
                        next_nodes.append((info, child_identifier, child_node))

            if len(next_nodes) == 0:
                break

            current_nodes = next_nodes
            current_depth += 1

        return root_info

    def collect_info(self, preserve_list=True):
        """
        Collect the information from the NodeSchemaInfo tree, and return it as nested dict.

        Parameters
        ----------

        preserve_list : bool
            If True, then lists are preserved. Otherwise, they are turned into dicts.
        """
        if preserve_list and isinstance(self.node, (list, tuple)) and self.info is None:
            info = [c_info for child in self.visible_children if len(c_info := child.collect_info(preserve_list)) > 0]
        else:
            info = {
                child.identifier: c_info
                for child in self.visible_children
                if len(c_info := child.collect_info(preserve_list)) > 0
            }

            if self.info is not None:
                info[self.key] = SchemaInfo(self.info, self.node)

        return info



import pytest

from asdf._commands import main


@pytest.mark.parametrize("type_string", ("numpy.ndarray", "asdf.tags.core.ndarray.NDArrayType"))
def test_search_array(capsys, test_data_path, type_string):
    """
    Test that both the lazy and non-lazy classes work for searches for arrays.
    """
    file_path = test_data_path / "ndarray_in_list0.asdf"

    assert main.main_from_args(["search", str(file_path), "--type", type_string]) == 0

    captured = capsys.readouterr()
    assert "list" in captured.out
    assert "a (NDArrayType)" in captured.out
    assert "history" not in captured.out


@pytest.mark.parametrize("limit_rows", (True, False))
@pytest.mark.parametrize("limit_cols", (True, False))
def test_search_rendering(capsys, test_data_path, limit_rows, limit_cols):
    """
    Test that search respects the rendering parameters common with info.
    """
    file_path = test_data_path / "ndarray_in_list0.asdf"

    max_rows = 5
    max_cols = 25

    args = ["search", str(file_path)]
    if limit_rows:
        args += ["--max-rows", str(max_rows)]
    if limit_cols:
        args += ["--max-cols", str(max_cols)]

    assert main.main_from_args(args) == 0

    captured = capsys.readouterr()
    lines = captured.out.splitlines()
    if limit_rows:
        assert len(lines) <= max_rows
    else:
        assert len(lines) > max_rows
    if limit_cols:
        assert all(len(line.rstrip()) <= max_cols for line in lines)
    else:
        assert any(len(line) > max_cols for line in lines)


# The test file contains a number of tags that aren't part of the standard.
# This is useful for these tests as we want to handle this case gracefully.
@pytest.mark.filterwarnings("ignore::asdf.exceptions.AsdfConversionWarning")
@pytest.mark.parametrize(
    "query_type, query, expected, not_expected",
    (
        ("key", "axes", "axes_names", "reference_frame"),
        ("key", "^axes_names$", "axes_names", "axes_order"),
        ("value", "ICRS", "ICRS", "FK4"),
        ("value", '["lon", "lat"]', "axes_names", "reference_frame"),
        ("value", '{"type": "ICRS"}', "reference_frame", "axes_names"),
        ("type", "int", "axes_order", "axes_names"),
        ("type", "float", "reference_frame", "axes_names"),
        ("type", "dict", "reference_frame", "axes_names"),
        ("type", "list", "axes_names", "asdf_library"),
    ),
)
def test_search_command(capsys, test_data_path, query_type, query, expected, not_expected):
    """
    Test that for a query and query_type search returns the expected value and doesn't
    return the not_expected value. These are all tailored to this particular file.
    """
    file_path = test_data_path / "frames0.asdf"

    assert main.main_from_args(["search", str(file_path), f"--{query_type}", query]) == 0

    captured = capsys.readouterr()
    assert expected in captured.out
    assert not_expected not in captured.out



from yaml.representer import RepresenterError

from asdf._jsonschema import ValidationError

__all__ = [
    "AsdfConversionWarning",
    "AsdfDeprecationWarning",
    "AsdfLazyReferenceError",
    "AsdfManifestURIMismatchWarning",
    "AsdfPackageVersionWarning",
    "AsdfProvisionalAPIWarning",
    "AsdfSerializationError",
    "AsdfWarning",
    "DelimiterNotFoundError",
    "ValidationError",
]


class AsdfWarning(Warning):
    """
    The base warning class from which all ASDF warnings should inherit.
    """


class AsdfDeprecationWarning(AsdfWarning, DeprecationWarning):
    """
    A warning class to indicate a deprecated feature.
    """


class AsdfConversionWarning(AsdfWarning):
    """
    Warning class used for failures to convert data into custom types.
    """


class AsdfBlockIndexWarning(AsdfWarning):
    """
    Warning class to indicate that a file was read with an invalid block index
    """


class DelimiterNotFoundError(ValueError):
    """
    Indicates that a delimiter was not found when reading or
    seeking through a file.
    """


class AsdfProvisionalAPIWarning(AsdfWarning, FutureWarning):
    """
    Used for provisional features where breaking API changes might be
    introduced at any point (including minor releases). These features
    are likely to be added in a future ASDF version. However, Use of
    provisional features is highly discouraged for production code.
    """


class AsdfPackageVersionWarning(AsdfWarning):
    """
    A warning indicating a package version mismatch
    """


class AsdfManifestURIMismatchWarning(AsdfWarning):
    """
    A warning indicaing that an extension registered with a manifest
    contains a id that does not match the uri of the manifest.
    """


class AsdfLazyReferenceError(ReferenceError):
    """
    Indicates that the lazy tree node failed to resolve a reference
    to an AsdfFile instance. This likely means the AsdfFile was garbage
    collected and you may need to update your code to keep the AsdfFile
    in memory (by keeping a reference).
    """


class AsdfSerializationError(RepresenterError):
    """
    An object failed serialization by asdf and by yaml. This likely indicates
    that the object does not have a supporting asdf Converter and needs to
    be manually converted to a supported type.
    """



import io
import os

import numpy as np
import pytest
from numpy.testing import assert_array_equal

import asdf
from asdf import Stream, generic_io


def test_stream():
    buff = io.BytesIO()

    tree = {"stream": Stream([6, 2], np.float64)}

    ff = asdf.AsdfFile(tree)
    ff.write_to(buff)
    for i in range(100):
        buff.write(np.array([i] * 12, np.float64).tobytes())

    buff.seek(0)

    with asdf.open(buff) as ff:
        assert len(ff._blocks.blocks) == 1
        assert ff.tree["stream"].shape == (100, 6, 2)
        for i, row in enumerate(ff.tree["stream"]):
            assert np.all(row == i)


def test_stream_write_nothing():
    """
    Test that if you write nothing, you get a zero-length array
    """

    buff = io.BytesIO()

    tree = {"stream": Stream([6, 2], np.float64)}

    ff = asdf.AsdfFile(tree)
    ff.write_to(buff)

    buff.seek(0)

    with asdf.open(buff) as ff:
        assert len(ff._blocks.blocks) == 1
        assert ff.tree["stream"].shape == (0, 6, 2)


def test_stream_twice():
    """
    Test that if you write nothing, you get a zero-length array
    """

    buff = io.BytesIO()

    tree = {"stream": Stream([6, 2], np.uint8), "stream2": Stream([12, 2], np.uint8)}

    ff = asdf.AsdfFile(tree)
    ff.write_to(buff)
    for i in range(100):
        buff.write(np.array([i] * 12, np.uint8).tobytes())

    buff.seek(0)

    ff = asdf.open(buff)
    assert len(ff._blocks.blocks) == 1
    assert ff.tree["stream"].shape == (100, 6, 2)
    assert ff.tree["stream2"].shape == (50, 12, 2)


def test_stream_with_nonstream():
    buff = io.BytesIO()

    tree = {"nonstream": np.array([1, 2, 3, 4], np.int64), "stream": Stream([6, 2], np.float64)}

    ff = asdf.AsdfFile(tree)
    # Since we're testing with small arrays, force this array to be stored in
    # an internal block rather than letting it be automatically put inline.
    ff.set_array_storage(ff["nonstream"], "internal")
    ff.write_to(buff)
    for i in range(100):
        buff.write(np.array([i] * 12, np.float64).tobytes())

    buff.seek(0)

    with asdf.open(buff) as ff:
        assert len(ff._blocks.blocks) == 2
        assert_array_equal(ff.tree["nonstream"], np.array([1, 2, 3, 4], np.int64))
        assert ff.tree["stream"].shape == (100, 6, 2)
        for i, row in enumerate(ff.tree["stream"]):
            assert np.all(row == i)


def test_stream_real_file(tmp_path):
    path = os.path.join(str(tmp_path), "test.asdf")

    tree = {"nonstream": np.array([1, 2, 3, 4], np.int64), "stream": Stream([6, 2], np.float64)}

    with open(path, "wb") as fd:
        ff = asdf.AsdfFile(tree)
        # Since we're testing with small arrays, force this array to be stored
        # in an internal block rather than letting it be automatically put
        # inline.
        ff.set_array_storage(ff["nonstream"], "internal")
        ff.write_to(fd)
        for i in range(100):
            fd.write(np.array([i] * 12, np.float64).tobytes())

    with asdf.open(path) as ff:
        assert len(ff._blocks.blocks) == 2
        assert_array_equal(ff.tree["nonstream"], np.array([1, 2, 3, 4], np.int64))
        assert ff.tree["stream"].shape == (100, 6, 2)
        for i, row in enumerate(ff.tree["stream"]):
            assert np.all(row == i)


def test_stream_to_stream():
    tree = {"nonstream": np.array([1, 2, 3, 4], np.int64), "stream": Stream([6, 2], np.float64)}

    buff = io.BytesIO()
    fd = generic_io.OutputStream(buff)

    ff = asdf.AsdfFile(tree)
    ff.write_to(fd)
    for i in range(100):
        fd.write(np.array([i] * 12, np.float64).tobytes())

    buff.seek(0)

    with asdf.open(generic_io.InputStream(buff, "r")) as ff:
        assert len(ff._blocks.blocks) == 2
        assert_array_equal(ff.tree["nonstream"], np.array([1, 2, 3, 4], np.int64))
        assert ff.tree["stream"].shape == (100, 6, 2)
        for i, row in enumerate(ff.tree["stream"]):
            assert np.all(row == i)


def test_array_to_stream(tmp_path):
    tree = {
        "stream": np.array([1, 2, 3, 4], np.int64),
    }

    buff = io.BytesIO()
    ff = asdf.AsdfFile(tree)
    ff.set_array_storage(tree["stream"], "streamed")
    ff.write_to(buff)
    buff.write(np.array([5, 6, 7, 8], np.int64).tobytes())

    buff.seek(0)
    ff = asdf.open(generic_io.InputStream(buff))
    assert_array_equal(ff.tree["stream"], [1, 2, 3, 4, 5, 6, 7, 8])
    buff.seek(0)
    ff2 = asdf.AsdfFile(ff)
    ff2.set_array_storage(ff2["stream"], "streamed")
    ff2.write_to(buff)
    assert b"shape: ['*']" in buff.getvalue()

    with open(os.path.join(str(tmp_path), "test.asdf"), "wb") as fd:
        ff = asdf.AsdfFile(tree)
        ff.set_array_storage(tree["stream"], "streamed")
        ff.write_to(fd)
        fd.write(np.array([5, 6, 7, 8], np.int64).tobytes())

    with asdf.open(os.path.join(str(tmp_path), "test.asdf")) as ff:
        assert_array_equal(ff.tree["stream"], [1, 2, 3, 4, 5, 6, 7, 8])
        ff2 = asdf.AsdfFile(ff)
        ff2.write_to(buff)
        assert b"shape: ['*']" in buff.getvalue()


def test_too_many_streams():
    tree = {"stream1": np.array([1, 2, 3, 4], np.int64), "stream2": np.array([1, 2, 3, 4], np.int64)}

    ff = asdf.AsdfFile(tree)
    ff.set_array_storage(tree["stream1"], "streamed")
    with pytest.raises(ValueError, match=r"Can not add second streaming block"):
        ff.set_array_storage(tree["stream2"], "streamed")


def test_stream_repr_and_str():
    tree = {"stream": Stream([16], np.int64)}

    ff = asdf.AsdfFile(tree)
    repr(ff.tree["stream"])
    str(ff.tree["stream"])



import io
import os

import numpy as np
import pytest
from numpy.testing import assert_array_equal

import asdf
from asdf import reference
from asdf.tags.core import ndarray

from ._helpers import assert_tree_match


def test_external_reference(tmp_path):
    exttree = {
        "cool_stuff": {"a": np.array([0, 1, 2], float), "b": np.array([3, 4, 5], float)},
        "list_of_stuff": ["foobar", 42, np.array([7, 8, 9], float)],
    }
    external_path = os.path.join(str(tmp_path), "external.asdf")
    ext = asdf.AsdfFile(exttree)
    # Since we're testing with small arrays, force all arrays to be stored
    # in internal blocks rather than letting some of them be automatically put
    # inline.
    ext.write_to(external_path, all_array_storage="internal")

    external_path = os.path.join(str(tmp_path), "external2.asdf")
    ff = asdf.AsdfFile(exttree)
    ff.write_to(external_path, all_array_storage="internal")

    tree = {
        # The special name "data" here must be an array.  This is
        # included so that such validation can be ignored when we just
        # have a "$ref".
        "data": {"$ref": "external.asdf#/cool_stuff/a"},
        "science_data": {"$ref": "external.asdf#/cool_stuff/a"},
        "science_data2": {"$ref": "external2.asdf#/cool_stuff/a"},
        "foobar": {
            "$ref": "external.asdf#/list_of_stuff/0",
        },
        "answer": {"$ref": "external.asdf#/list_of_stuff/1"},
        "array": {
            "$ref": "external.asdf#/list_of_stuff/2",
        },
        "whole_thing": {"$ref": "external.asdf#"},
        "myself": {
            "$ref": "#",
        },
        "internal": {"$ref": "#science_data"},
    }

    def do_asserts(ff):
        assert "unloaded" in repr(ff.tree["science_data"])
        assert "unloaded" in str(ff.tree["science_data"])
        assert len(ff._external_asdf_by_uri) == 0

        assert_array_equal(ff.tree["science_data"], exttree["cool_stuff"]["a"])
        assert len(ff._external_asdf_by_uri) == 1

        assert_array_equal(ff.tree["science_data2"], exttree["cool_stuff"]["a"])
        assert len(ff._external_asdf_by_uri) == 2

        assert ff.tree["foobar"]() == "foobar"
        assert ff.tree["answer"]() == 42
        assert_array_equal(ff.tree["array"], exttree["list_of_stuff"][2])

        assert_tree_match(ff.tree["whole_thing"](), exttree)

        assert_array_equal(ff.tree["whole_thing"]["cool_stuff"]["a"], exttree["cool_stuff"]["a"])

        assert_array_equal(ff.tree["myself"]["science_data"], exttree["cool_stuff"]["a"])
        # Make sure that referencing oneself doesn't make another call
        # to disk.
        assert len(ff._external_asdf_by_uri) == 2

        assert_array_equal(ff.tree["internal"], exttree["cool_stuff"]["a"])

    with asdf.AsdfFile({}, uri=(tmp_path / "main.asdf").as_uri()) as ff:
        ff.tree = tree
        ff.find_references()
        do_asserts(ff)

        internal_path = os.path.join(str(tmp_path), "main.asdf")
        ff.write_to(internal_path)

    with asdf.open(internal_path) as ff:
        ff.find_references()
        do_asserts(ff)

    with asdf.open(internal_path) as ff:
        ff.find_references()
        assert len(ff._external_asdf_by_uri) == 0
        ff.resolve_references()
        assert len(ff._external_asdf_by_uri) == 2

        assert isinstance(ff.tree["data"], ndarray.NDArrayType)
        assert isinstance(ff.tree["science_data"], ndarray.NDArrayType)

        assert_array_equal(ff.tree["science_data"], exttree["cool_stuff"]["a"])
        assert_array_equal(ff.tree["science_data2"], exttree["cool_stuff"]["a"])

        assert ff.tree["foobar"] == "foobar"
        assert ff.tree["answer"] == 42
        assert_array_equal(ff.tree["array"], exttree["list_of_stuff"][2])

        assert_tree_match(ff.tree["whole_thing"], exttree)

        assert_array_equal(ff.tree["whole_thing"]["cool_stuff"]["a"], exttree["cool_stuff"]["a"])

        assert_array_equal(ff.tree["myself"]["science_data"], exttree["cool_stuff"]["a"])

        assert_array_equal(ff.tree["internal"], exttree["cool_stuff"]["a"])


@pytest.mark.remote_data()
def test_external_reference_invalid(tmp_path):
    tree = {"foo": {"$ref": "fail.asdf"}}

    ff = asdf.AsdfFile()
    ff.tree = tree
    ff.find_references()
    with pytest.raises(ValueError, match=r"Resolved to relative URL"):
        ff.resolve_references()

    ff = asdf.AsdfFile({}, uri="http://httpstat.us/404")
    ff.tree = tree
    ff.find_references()
    msg = r"[HTTP Error 404: Not Found, HTTP Error 502: Bad Gateway]"  # if httpstat.us is down 502 is returned.
    with pytest.raises(IOError, match=msg):
        ff.resolve_references()

    ff = asdf.AsdfFile({}, uri=(tmp_path / "main.asdf").as_uri())
    ff.tree = tree
    ff.find_references()
    with pytest.raises(IOError, match=r"No such file or directory: .*"):
        ff.resolve_references()


def test_external_reference_invalid_fragment(tmp_path):
    exttree = {"list_of_stuff": ["foobar", 42, np.array([7, 8, 9], float)]}
    external_path = os.path.join(str(tmp_path), "external.asdf")
    ff = asdf.AsdfFile(exttree)
    ff.write_to(external_path)

    tree = {"foo": {"$ref": "external.asdf#/list_of_stuff/a"}}

    with asdf.AsdfFile({}, uri=(tmp_path / "main.asdf").as_uri()) as ff:
        ff.tree = tree
        ff.find_references()
        with pytest.raises(ValueError, match=r"Unresolvable reference: .*"):
            ff.resolve_references()

    tree = {"foo": {"$ref": "external.asdf#/list_of_stuff/3"}}

    with asdf.AsdfFile({}, uri=(tmp_path / "main.asdf").as_uri()) as ff:
        ff.tree = tree
        ff.find_references()
        with pytest.raises(ValueError, match=r"Unresolvable reference: .*"):
            ff.resolve_references()


def test_make_reference(tmp_path):
    exttree = {
        # Include some ~ and / in the name to make sure that escaping
        # is working correctly
        "f~o~o/": {"a": np.array([0, 1, 2], float), "b": np.array([3, 4, 5], float)},
    }
    external_path = os.path.join(str(tmp_path), "external.asdf")
    ext = asdf.AsdfFile(exttree)
    ext.write_to(external_path)

    with asdf.open(external_path) as ext:
        ff = asdf.AsdfFile()
        ff.tree["ref"] = ext.make_reference(["f~o~o/", "a"])
        assert_array_equal(ff.tree["ref"], ext.tree["f~o~o/"]["a"])

        ff.write_to(os.path.join(str(tmp_path), "source.asdf"))

    with (asdf.open(os.path.join(str(tmp_path), "source.asdf")) as ff,):
        ff.find_references()
        assert ff.tree["ref"]._uri == "external.asdf#f~0o~0o~1/a"


def test_internal_reference(tmp_path):
    testfile = tmp_path / "test.asdf"

    tree = {"foo": 2, "bar": {"$ref": "#"}}

    ff = asdf.AsdfFile(tree)
    ff.find_references()
    assert isinstance(ff.tree["bar"], reference.Reference)
    ff.resolve_references()
    assert ff.tree["bar"]["foo"] == 2

    tree = {"foo": 2}
    ff = asdf.AsdfFile(tree, uri=testfile.as_uri())
    ff.tree["bar"] = ff.make_reference([])
    buff = io.BytesIO()
    ff.write_to(buff)
    buff.seek(0)
    ff = asdf.AsdfFile()
    assert b"{$ref: ''}" in buff.getvalue()


def test_implicit_internal_reference(tmp_path, with_lazy_tree):
    target = {"foo": "bar"}
    nested_in_dict = {"target": target}
    nested_in_list = [target]
    tree = {"target": target, "nested_in_dict": nested_in_dict, "nested_in_list": nested_in_list}

    assert tree["target"] is tree["nested_in_dict"]["target"]
    assert tree["target"] is tree["nested_in_list"][0]

    af = asdf.AsdfFile(tree)

    assert af["target"] is af["nested_in_dict"]["target"]
    assert af["target"] is af["nested_in_list"][0]

    output_path = os.path.join(str(tmp_path), "test.asdf")
    af.write_to(output_path)
    with asdf.open(output_path) as af:
        assert af["target"] is af["nested_in_dict"]["target"]
        assert af["target"] is af["nested_in_list"][0]



import warnings
import weakref

from asdf import constants
from asdf.exceptions import AsdfBlockIndexWarning, AsdfWarning, DelimiterNotFoundError

from . import io as bio
from .exceptions import BlockIndexError


class ReadBlock:
    """
    Represents an ASDF block read from a file.
    """

    def __init__(self, offset, fd, memmap, lazy_load, validate_checksum, header=None, data_offset=None, data=None):
        self.offset = offset  # after block magic bytes
        self._fd = weakref.ref(fd)
        self._header = header
        self.data_offset = data_offset
        self._data = data
        self._cached_data = None
        self.memmap = memmap
        self.lazy_load = lazy_load
        self.validate_checksum = validate_checksum
        if not lazy_load:
            self.load()

    def close(self):
        self._cached_data = None

    @property
    def loaded(self):
        return self._data is not None

    def load(self):
        """
        Load the block data (if it is not already loaded).

        Raises
        ------
        OSError
            If attempting to load from a closed file.
        """
        if self.loaded:
            return
        fd = self._fd()
        if fd is None or fd.is_closed():
            msg = "Attempt to load block from closed file"
            raise OSError(msg)
        position = fd.tell()
        _, self._header, self.data_offset, self._data = bio.read_block(
            fd, offset=self.offset, memmap=self.memmap, lazy_load=self.lazy_load
        )
        fd.seek(position)

    @property
    def data(self):
        """
        Read, parse and return data for an ASDF block.

        Returns
        -------
        data : ndarray
            A one-dimensional ndarray of dypte uint8 read from an ASDF block

        Raises
        ------
        ValueError
            If the header checksum does not match the checksum of the data
            and validate_checksums was set to True.
        """
        if not self.loaded:
            self.load()
        if callable(self._data):
            data = self._data()
        else:
            data = self._data
        if self.validate_checksum:
            checksum = bio.calculate_block_checksum(data)
            if not self._header["flags"] & constants.BLOCK_FLAG_STREAMED and checksum != self._header["checksum"]:
                msg = f"Block at {self.offset} does not match given checksum"
                raise ValueError(msg)
            # only validate data the first time it's read
            self.validate_checksum = False
        return data

    @property
    def cached_data(self):
        """
        Return cached data for an ASDF block.

        The first time this is called it may read data from the file
        (if lazy loaded). Subsequent calls will return the same
        ndarray.
        """
        if self._cached_data is None:
            self._cached_data = self.data
        return self._cached_data

    @property
    def header(self):
        """
        Get the block header. For a lazy loaded block the first time
        this is called the header will be read from the file and
        cached.

        Returns
        -------
        header : dict
            Dictionary containing the read ASDF header.
        """
        if not self.loaded:
            self.load()
        return self._header


def _read_blocks_serially(fd, memmap=False, lazy_load=False, validate_checksums=False, after_magic=False):
    """
    Read blocks serially from a file without looking for a block index.

    For parameter and return value descriptions see `read_blocks`.
    """
    blocks = []
    magic_len = len(constants.BLOCK_MAGIC)

    if not after_magic:
        # seek until the first magic is found
        try:
            fd.seek_until(b"(" + constants.BLOCK_MAGIC + b")", magic_len)
        except DelimiterNotFoundError:
            return blocks
        after_magic = True

    buff = constants.BLOCK_MAGIC
    while buff == constants.BLOCK_MAGIC:
        # read the block
        offset, header, data_offset, data = bio.read_block(fd, memmap=memmap, lazy_load=lazy_load)
        blocks.append(
            ReadBlock(
                offset, fd, memmap, lazy_load, validate_checksums, header=header, data_offset=data_offset, data=data
            )
        )
        if blocks[-1].header["flags"] & constants.BLOCK_FLAG_STREAMED:
            # a file can only have 1 streamed block and it must be at the end so we
            # can stop looking for more blocks
            return blocks

        # check for the next block
        buff = fd.read(magic_len)

    # check remaining bytes
    if buff == constants.INDEX_HEADER[: len(buff)]:
        # remaining bytes are the start of the block index
        return blocks
    if buff == b"\0" * len(buff):
        # remaining bytes are null
        return blocks
    msg = f"Read invalid bytes {buff!r} after blocks, your file might be corrupt"
    warnings.warn(msg, AsdfWarning)
    return blocks


def read_blocks(fd, memmap=False, lazy_load=False, validate_checksums=False, after_magic=False):
    """
    Read a sequence of ASDF blocks from a file.

    If the file is seekable (and lazy_load is False) an attempt will
    made to find, read and parse a block index. If this fails, the
    blocks will be read serially. If parsing the block index
    succeeds, the first first and last blocks will be read (to
    confirm that those portions of the index are correct). All
    other blocks will not be read until they are accessed.

    Parameters
    ----------
    fd : file or generic_io.GenericIO
        File to read. Reading will start at the current position.

    memmap : bool, optional, default False
        If true, memory map block data.

    lazy_load : bool, optional, default False
        If true, block data will be a callable that when executed
        will return the block data. See the ``lazy_load`` argument
        to ``asdf._block.io.read_block`` for more details.

    validate_checksums : bool, optional, default False
        When reading blocks compute the block data checksum and
        compare it to the checksum read from the block header.
        Note that this comparison will occur when the data is
        accessed if ``lazy_load`` was set to True.

    after_magic : bool, optional, default False
        If True don't expect block magic bytes for the first block
        read from the file.

    Returns
    -------

    read_blocks : list of ReadBlock
        A list of ReadBlock instances.

    Raises
    ------
    OSError
        Invalid bytes encountered while reading blocks.

    ValueError
        A read block has an invalid checksum.
    """
    if not lazy_load or not fd.seekable():
        # load all blocks serially
        return _read_blocks_serially(fd, memmap, lazy_load, validate_checksums, after_magic)

    # try to find block index
    starting_offset = fd.tell()
    index_offset = bio.find_block_index(fd, starting_offset)
    if index_offset is None:
        # if failed, load all blocks serially
        fd.seek(starting_offset)
        return _read_blocks_serially(fd, memmap, lazy_load, validate_checksums, after_magic)

    # setup empty blocks
    try:
        block_index = bio.read_block_index(fd, index_offset)
    except BlockIndexError as e:
        # failed to read block index, fall back to serial reading
        msg = f"Failed to read block index, falling back to serial reading: {e!s}"
        warnings.warn(msg, AsdfBlockIndexWarning)
        fd.seek(starting_offset)
        return _read_blocks_serially(fd, memmap, lazy_load, validate_checksums, after_magic)
    # skip magic for each block
    magic_len = len(constants.BLOCK_MAGIC)
    blocks = [ReadBlock(offset + magic_len, fd, memmap, lazy_load, validate_checksums) for offset in block_index]
    try:
        # load first and last blocks to check if the index looks correct
        for index in (0, -1):
            fd.seek(block_index[index])
            buff = fd.read(magic_len)
            if buff != constants.BLOCK_MAGIC:
                msg = "Invalid block magic"
                raise OSError(msg)
            blocks[index].load()
    except (OSError, ValueError) as e:
        msg = f"Invalid block index contents for block {index}, falling back to serial reading: {e!s}"
        warnings.warn(msg, AsdfBlockIndexWarning)
        fd.seek(starting_offset)
        return _read_blocks_serially(fd, memmap, lazy_load, after_magic)
    return blocks



import copy
import gc

from asdf._block.key import Key


# a blank class for testing
class Foo:
    pass


def test_unique_per_object():
    seen = set()
    for _i in range(10):
        bk = Key(Foo())
        assert bk not in seen
        seen.add(bk)


def test_unique_same_object():
    seen = set()
    f = Foo()
    for _i in range(10):
        bk = Key(f)
        assert bk not in seen
        seen.add(bk)


def test_matches_obj():
    f = Foo()
    bk = Key(f)
    assert bk._matches_object(f)


def test_undefined_no_match():
    bk = Key()
    assert not bk._matches_object(Foo())


def test_is_valid():
    f = Foo()
    bk = Key(f)
    assert bk._is_valid()
    del f
    gc.collect(2)
    assert not bk._is_valid()


def test_same_class():
    f = Foo()
    bk = Key(f)
    del f
    gc.collect(2)
    f2 = Foo()
    assert not bk._is_valid()
    assert not bk._matches_object(f2)


def test_undefined():
    k = Key()
    assert not k._is_valid()


def test_equal():
    key_value = 42
    f = Foo()
    k1 = Key(f, key_value)
    k2 = Key(f, key_value)
    assert k1 == k2


def test_key_mismatch_not_equal():
    f = Foo()
    k1 = Key(f)
    k2 = Key(f)
    assert k1 != k2


def test_obj_not_equal():
    f = Foo()
    k = Key(f)
    assert k != f


def test_undefined_not_equal():
    key_value = 42
    k1 = Key(_key=key_value)
    k2 = Key(_key=key_value)
    assert k1 != k2


def test_deleted_object_not_equal():
    key_value = 42
    f = Foo()
    k1 = Key(f, key_value)
    k2 = Key(f, key_value)
    del f
    gc.collect(2)
    assert k1 != k2


def test_copy():
    f = Foo()
    k1 = Key(f)
    k2 = copy.copy(k1)
    assert k1 == k2


def test_copy_undefined_not_equal():
    k1 = Key()
    k2 = copy.copy(k1)
    assert k1 != k2


def test_copy_deleted_object_not_equal():
    f = Foo()
    k1 = Key(f)
    k2 = copy.copy(k1)
    del f
    gc.collect(2)
    assert k1 != k2



from asdf.extension import Converter


class AsdfObjectConverter(Converter):
    # Since AsdfObject is just a dict, we're able to use the same converter
    # for both tag versions.
    tags = [
        "tag:stsci.edu:asdf/core/asdf-1.0.0",
        "tag:stsci.edu:asdf/core/asdf-1.1.0",
    ]
    types = ["asdf.tags.core.AsdfObject"]

    def to_yaml_tree(self, obj, tag, ctx):
        return dict(obj)

    def from_yaml_tree(self, node, tag, ctx):
        from asdf.tags.core import AsdfObject

        return AsdfObject(node)


class ExtensionMetadataConverter(Converter):
    tags = ["tag:stsci.edu:asdf/core/extension_metadata-1.0.0"]
    types = ["asdf.tags.core.ExtensionMetadata"]

    def to_yaml_tree(self, obj, tag, ctx):
        return dict(obj)

    def from_yaml_tree(self, node, tag, ctx):
        from asdf.tags.core import ExtensionMetadata

        return ExtensionMetadata(node)


class HistoryEntryConverter(Converter):
    tags = ["tag:stsci.edu:asdf/core/history_entry-1.0.0"]
    types = ["asdf.tags.core.HistoryEntry"]

    def to_yaml_tree(self, obj, tag, ctx):
        return dict(obj)

    def from_yaml_tree(self, node, tag, ctx):
        from asdf.tags.core import HistoryEntry

        return HistoryEntry(node)


class SoftwareConverter(Converter):
    tags = ["tag:stsci.edu:asdf/core/software-1.0.0"]
    types = ["asdf.tags.core.Software"]

    def to_yaml_tree(self, obj, tag, ctx):
        return dict(obj)

    def from_yaml_tree(self, node, tag, ctx):
        from asdf.tags.core import Software

        return Software(node)


class SubclassMetadataConverter(Converter):
    tags = ["tag:stsci.edu:asdf/core/subclass_metadata-1.0.0"]
    types = ["asdf.tags.core.SubclassMetadata"]

    def to_yaml_tree(self, obj, tag, ctx):
        return dict(obj)

    def from_yaml_tree(self, node, tag, ctx):
        from asdf.tags.core import SubclassMetadata

        return SubclassMetadata(node)



import io
import sys

import pytest

import asdf
from asdf._commands import diff, main


@pytest.fixture(autouse=True)
def force_isatty(monkeypatch):
    def _isatty():
        return True

    monkeypatch.setattr(sys.stdout, "isatty", _isatty)
    yield


def _assert_diffs_equal(test_data_path, filenames, result_file, minimal=False, ignore=None):
    iostream = io.StringIO()

    file_paths = [test_data_path / name for name in filenames]
    diff(file_paths, minimal=minimal, iostream=iostream, ignore=ignore)
    iostream.seek(0)

    result_path = test_data_path / result_file
    with open(result_path) as handle:
        assert handle.read() == iostream.read()


def test_diff(test_data_path):
    filenames = ["frames0.asdf", "frames1.asdf"]
    result_file = "frames.diff"
    _assert_diffs_equal(test_data_path, filenames, result_file, minimal=False)


def test_diff_minimal(test_data_path):
    filenames = ["frames0.asdf", "frames1.asdf"]
    result_file = "frames_minimal.diff"
    _assert_diffs_equal(test_data_path, filenames, result_file, minimal=True)


@pytest.mark.parametrize(
    ("result_file", "ignore"),
    [
        ("frames_ignore_asdf_library.diff", ["asdf_library"]),
        ("frames_ignore_reference_frame.diff", ["frames[*].reference_frame"]),
        ("frames_ignore_both.diff", ["asdf_library", "frames[*].reference_frame"]),
    ],
)
def test_diff_ignore(test_data_path, result_file, ignore):
    filenames = ["frames0.asdf", "frames1.asdf"]
    _assert_diffs_equal(test_data_path, filenames, result_file, minimal=False, ignore=ignore)


def test_diff_ndarray(test_data_path):
    filenames = ["ndarray0.asdf", "ndarray1.asdf"]
    result_file = "ndarrays.diff"
    _assert_diffs_equal(test_data_path, filenames, result_file, minimal=False)


def test_diff_ndarray_in_list(test_data_path):
    filenames = ["ndarray_in_list0.asdf", "ndarray_in_list1.asdf"]
    result_file = "ndarray_in_list.diff"
    _assert_diffs_equal(test_data_path, filenames, result_file, minimal=False)


def test_diff_block(test_data_path):
    filenames = ["block0.asdf", "block1.asdf"]
    result_file = "blocks.diff"
    _assert_diffs_equal(test_data_path, filenames, result_file, minimal=False)


def test_diff_simple_inline_array(test_data_path):
    filenames = ["simple_inline_array0.asdf", "simple_inline_array1.asdf"]
    result_file = "simple_inline_array.diff"
    _assert_diffs_equal(test_data_path, filenames, result_file, minimal=False)


@pytest.mark.filterwarnings("ignore:unclosed file .*")
def test_file_not_found(test_data_path):
    # Try to open files that exist but are not valid asdf
    filenames = ["frames.diff", "blocks.diff"]
    with pytest.raises(
        RuntimeError,
        match=r"Does not appear to be a ASDF file.",
    ):
        diff([test_data_path / name for name in filenames], False)


def test_diff_command(test_data_path):
    filenames = ["frames0.asdf", "frames1.asdf"]
    path_strings = [str(test_data_path / name) for name in filenames]

    assert main.main_from_args(["diff", *path_strings]) == 0


def test_recursive_diff(tmp_path):
    fn0 = tmp_path / "a.asdf"
    fn1 = tmp_path / "b.asdf"

    t0 = {}
    t0["a"] = t0

    t1 = {"a": []}
    t1["a"].append(t1)

    asdf.dump(t0, fn0)
    asdf.dump(t1, fn1)

    assert main.main_from_args(["diff", str(fn0), str(fn1)]) == 0


def test_int_key_diff(tmp_path):
    fn0 = tmp_path / "a.asdf"
    fn1 = tmp_path / "b.asdf"

    t0 = {1: "a"}
    t1 = {2: "b"}

    asdf.dump(t0, fn0)
    asdf.dump(t1, fn1)

    assert main.main_from_args(["diff", str(fn0), str(fn1)]) == 0
